#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""semopy 2.0 model module without random effects."""
from .utils import chol_inv, chol_inv2, cov, delete_mx, chol
from statsmodels.stats.correlation_tools import cov_nearest
from itertools import combinations, chain
from .constraints import parse_constraint
from collections import defaultdict
from dataclasses import dataclass
from .model_base import ModelBase
from .polycorr import hetcor
from .solver import Solver
from . import startingvalues
import pandas as pd
import numpy as np
import logging


class Model(ModelBase):
    """Model class without mean-structure."""

    symb_starting_values = 'START'
    symb_bound_parameters = 'BOUND'
    symb_constraint = 'CONSTRAINT'

    # Make sure that approriate function build_matrixname() is present.
    matrices_names = 'beta', 'lambda', 'psi', 'theta'

    @dataclass
    class ParameterLoc:
        """Structure for keeping track of parameter's location in matrices."""

        __slots__ = ['matrix', 'indices', 'symmetric']

        matrix: np.ndarray
        indices: tuple
        symmetric: bool

    @dataclass
    class Parameter:
        """Structure for basic parameter info used internally in Model."""

        __slots__ = ['start', 'active', 'bound', 'locations']

        start: float
        active: bool
        bound: tuple
        locations: list

    def __init__(self, description: str, mimic_lavaan=False, baseline=False,
                 cov_diag=False):
        """
        Instantiate Model without mean-structure.

        Parameters
        ----------
        description : str
            Model description in semopy syntax.
        mimic_lavaan: bool
            If True, output variables are correlated and not conceptually
            identical to indicators. lavaan treats them that way, but it's
            less computationally effective. The default is False.
        baseline : bool
            If True, the model will be set to baseline model.
            Baseline model here is an independence model where all variables
            are considered to be independent with zero covariance. Only
            variances are estimated. The default is False.
        cov_diag : bool
            If cov_diag is True, then there are no covariances parametrised
            unless explicitly specified. The default is False.

        Returns
        -------
        None.

        """
        self.mimic_lavaan = mimic_lavaan
        self.parameters = dict()
        self.n_param_reg = 0
        self.n_param_cov = 0
        self.baseline = baseline
        self.constraints = list()
        self.cov_diag = cov_diag
        dops = self.dict_operations
        dops[self.symb_starting_values] = self.operation_start
        dops[self.symb_bound_parameters] = self.operation_bound
        dops[self.symb_constraint] = self.operation_constraint
        self.objectives = {'MLW': (self.obj_mlw, self.grad_mlw),
                           'ULS': (self.obj_uls, self.grad_uls),
                           'GLS': (self.obj_gls, self.grad_gls),
                           'WLS': (self.obj_wls, self.grad_wls),
                           'DWLS': (self.obj_dwls, self.grad_dwls),
                           'FIML': (self.obj_fiml, self.grad_fiml)}
        super().__init__(description)

    def add_param(self, name: str, active: bool, start: float, bound: tuple,
                  matrix: np.ndarray, indices: tuple, symmetric: bool):
        """
        Add parameter/update parameter locations in semopy matrices.

        If name is not present in self.parameters, then just locations will be
        updated. Otherwise, a new Parameter instance is added to
        self.parameters.
        Parameters
        ----------
        name : str
            Name of parameter.
        active : bool
            Is the parameter "active", i.e. is it subject to further
            optimization.
        start : float
            Starting value of parameter. If parameter is not active, then it is
            a fixing value.
        bound : tuple
            Bound constraints on parameter (a, b). None is treated as infinity.
        matrix : np.ndarray
            Reference to matrix.
        indices : tuple
            Indices of parameter in the matrix.
        symmetric : bool
            Should be True if matrix is symmetric.

        Returns
        -------
        None.

        """
        loc = self.ParameterLoc(matrix=matrix, indices=indices,
                                symmetric=symmetric)
        if name in self.parameters:
            p = self.parameters[name]
            p.locations.append(loc)
        else:
            p = self.Parameter(active=active, start=start, bound=bound,
                               locations=[loc])
            self.parameters[name] = p

    def prepare_params(self):
        """
        Prepare structures for effective optimization routines.

        Returns
        -------
        None.

        """
        active_params = {name: param
                         for name, param in self.parameters.items()
                         if param.active}
        param_vals = [None] * len(active_params)
        diff_matrices = [None] * len(active_params)
        ranges = [[list(), list()] for _ in self.matrices]
        for i, (_, param) in enumerate(active_params.items()):
            param_vals[i] = param.start
            dm = [0] * len(self.matrices)
            for loc in param.locations:
                n = next(n for n in range(len(self.matrices))
                         if self.matrices[n] is loc.matrix)
                ranges[n][0].append(loc.indices)
                ranges[n][1].append(i)
                t = np.zeros_like(loc.matrix)
                t[loc.indices] = 1.0
                if loc.symmetric:
                    rind = loc.indices[::-1]
                    ranges[n][0].append(rind)
                    ranges[n][1].append(i)
                    t[rind] = 1.0
                dm[n] += t
            diff_matrices[i] = [m if type(m) is not int else None for m in dm]
        for rng in ranges:
            rng[0] = tuple(zip(*rng[0]))
        self.param_vals = np.array(param_vals)
        self.param_ranges = ranges
        self.mx_diffs = diff_matrices
        self.identity_c = np.identity(self.mx_beta.shape[0])
    

    def post_classification(self, effects: dict):
        """
        Procedure that is run just after classify_variables.

        Parameters
        -------
        effects : dict
            Maping opcode->values->rvalues->mutiplicator.

        Returns
        -------
        None.

        """
        self.finalize_variable_classification(effects)
        self.preprocess_effects(effects)
        self.setup_matrices()

    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply covariances to model.
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        cov = effects[self.symb_covariance]
        exo = self.vars['exogenous']
        obs_exo = set(self.vars['observed']) & exo
        for v in obs_exo:
            if v not in cov[v]:
                cov[v][v] = self.symb_starting_values
        for v in chain(self.vars['endogenous'], self.vars['latent']):
            if v not in cov[v]:
                cov[v][v] = None
        if not self.cov_diag:
            for a, b in combinations(obs_exo, 2):
                if a not in cov[b] and b not in cov[a]:
                    cov[a][b] = self.symb_starting_values
            exo_lat = self.vars['exogenous'] & self.vars['latent']
            for a, b in chain(combinations(self.vars['output'], 2),
                              combinations(exo_lat, 2)):
                if b not in cov[a] and a not in cov[b]:
                    cov[a][b] = None

    def finalize_variable_classification(self, effects: dict):
        """
        Finalize variable classification.

        Reorders variables for better visual fancyness and does extra
        model-specific variable respecification.

        Parameters
        -------
        effects : dict
            Maping opcode->values->rvalues->mutiplicator.

        Returns
        -------
        None.

        """
        if not self.mimic_lavaan:
            outputs = self.vars['output']
            self.vars['_output'] = outputs
            self.vars['output'] = {}
        else:
            inds = self.vars['indicator']
            # It's possible that user violates that classical requirement that
            # measurement variables must be outputs, here we take such case
            # in account by ANDing inds with outputs.
            self.vars['_output'] = inds & self.vars['output']
            self.vars['output'] -= inds
        inners = self.vars['all'] - self.vars['_output']
        self.vars['inner'] = inners
        # This is entirely for visual reasons only, one might replace sorted
        # with list as well.
        obs = self.vars['observed']
        inners = sorted(self.vars['latent'] & inners) + sorted(obs & inners)
        self.vars['inner'] = inners
        t = obs & self.vars['_output']
        self.vars['observed'] = sorted(t) + sorted(obs - t)

    def setup_matrices(self):
        """
        Initialize base matrix structures of the model.

        Returns
        -------
        None.

        """
        # I don't use dicts here as matrices structures and further ranges
        # structures will be used very often in optimsation procedure, and
        # key-retrieival process takes a toll on performance.
        # Please, don't rearrange matrices below, let the order stay the same.
        # It's necessary because I assumed fixed matrix positions in those
        # structures when coding linear algebra parts of the code.
        self.matrices = list()
        self.names = list()
        self.start_rules = list()
        for v in self.matrices_names:
            mx, names = getattr(self, f'build_{v}')()
            setattr(self, f'mx_{v}', mx)
            setattr(self, f'names_{v}', names)
            self.matrices.append(mx)
            self.names.append(names)
            try:
                self.start_rules.append(getattr(startingvalues, f'start_{v}'))
            except AttributeError:
                self.start_rules.append(getattr(self, f'start_{v}'))

    def build_beta(self):
        """
        Beta matrix contains relationships between all non-_output variables.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        names = self.vars['inner']
        n = len(names)
        mx = np.zeros((n, n))
        return mx, (names, names)

    def build_lambda(self):
        """
        Lambda matrix loads non-_output variables onto _output variables.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        obs = self.vars['observed']
        inner = self.vars['inner']
        row, col = obs, inner
        n, m = len(row), len(col)
        mx = np.zeros((n, m))
        for v in obs:
            if v in inner:
                i, j = obs.index(v), inner.index(v)
                mx[i, j] = 1.0
        return mx, (row, col)

    def build_psi(self):
        """
        Psi matrix is a covariance matrix for non-_output variables.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        names = self.vars['inner']
        n = len(names)
        mx = np.zeros((n, n))
        return mx, (names, names)

    def build_theta(self):
        """
        Theta matrix is a covariance matrix for _output variables.

        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        names = self.vars['observed']
        n = len(names)
        mx = np.zeros((n, n))
        return mx, (names, names)

    def effect_regression(self, items: dict):
        """
        Work through regression operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        if self.baseline:
            return
        outputs = self.vars['_output']
        for lv, rvs in items.items():
            lv_is_out = lv in outputs
            if lv_is_out:
                mx = self.mx_lambda
                rows, cols = self.names_lambda
            else:
                mx = self.mx_beta
                rows, cols = self.names_beta
            i = rows.index(lv)
            for rv, mult in rvs.items():
                name = None
                active = True
                try:
                    val = float(mult)
                    active = False
                except (TypeError, ValueError):
                    if mult is not None:
                        if mult == self.symb_starting_values:
                            active = False
                        else:
                            name = mult
                    val = None
                if name is None:
                    self.n_param_reg += 1
                    name = '_b%s' % self.n_param_reg
                j = cols.index(rv)
                ind = (i, j)
                self.add_param(name=name, matrix=mx, indices=ind, start=val,
                               active=active, symmetric=False,
                               bound=(None, None))

    def effect_measurement(self, items: dict):
        """
        Work through measurement operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Raises
        -------
        Exception
            Rises when indicator is misspecified and not observable.

        Returns
        -------
        None.

        """
        reverse_dict = defaultdict(dict)
        self.first_manifs = defaultdict(lambda: None)
        obs = self.vars['observed']
        for lat, inds in items.items():
            first = None
            lt = list()
            for ind, mult in inds.items():
                if ind not in obs:
                    logging.info(f'Manifest variables should be observed,\
                                   but {ind} appears to be latent.')
                try:
                    float(mult)
                    if first is None:
                        first = len(lt)
                except (TypeError, ValueError):
                    if mult == self.symb_starting_values and first is None:
                        first = len(lt)
                lt.append((ind, mult))
            if first is None:
                for i, (ind, mult) in enumerate(lt):
                    if mult is None:
                        first = i
                        break
                if first is None:
                    logging.warning('No fixed loadings for %s.', lat)
                else:
                    lt[first] = (ind, 1.0)
            for ind, mult in lt:
                reverse_dict[ind][lat] = mult
            if first is not None:
                self.first_manifs[lat] = lt[first][0]
        self.effect_regression(reverse_dict)

    def effect_covariance(self, items: dict):
        """
        Work through covariance operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        inners = self.vars['inner']
        lats = self.vars['latent']
        for lv, rvs in items.items():
            lv_is_inner = lv in inners
            for rv, mult in rvs.items():
                name = None
                try:
                    val = float(mult)
                    active = False
                except (TypeError, ValueError):
                    active = True
                    if mult is not None:
                        if mult != self.symb_starting_values:
                            name = mult
                        else:
                            active = False
                    val = None
                rv_is_inner = rv in inners
                if name is None:
                    self.n_param_cov += 1
                    name = '_c%s' % self.n_param_cov
                if lv_is_inner and rv_is_inner:
                    mx = self.mx_psi
                    rows, cols = self.names_psi
                else:
                    mx = self.mx_theta
                    rows, cols = self.names_theta
                    if lv_is_inner != rv_is_inner:
                        logging.info('Covariances between _outputs and \
                                     inner variables are not recommended.')
                i, j = rows.index(lv), cols.index(rv)
                ind = (i, j)
                if i == j:
                    if self.baseline and lv in lats:
                        continue
                    bound = (0, None)
                    symm = False
                else:
                    if self.baseline:
                        continue
                    bound = (None, None)
                    symm = True
                self.add_param(name, matrix=mx, indices=ind, start=val,
                               active=active, symmetric=symm, bound=bound)

    def inspect(self, mode='list', what='est', information='expected',
                std_est=False, se_robust=False):
        """
        Get fancy view of model parameters estimates.
    
        Parameters
        ----------
        model : str
            Model.
        mode : str, optional
            If 'list', pd.DataFrame with estimates and p-values is returned.
            If 'mx', a dictionary of matrices is returned. The default is 'list'.
        what : TYPE, optional
            Used only if mode == 'mx'. If 'est', matrices have estimated
            values. If 'start', matrices have starting values. If 'name', matrices
            have names inplace of their parameters. The default is 'est'.
        information : str
            If 'expected', expected Fisher information is used. Otherwise,
            observed information is employed. No use if mode is not 'list'.
            The default is 'expected'.
        std_est : bool
            If True, standardized coefficients are also returned as Std. Ests
            col. If it is 'lv', then output variables are not standardized.
            The default is False.
        se_robust : bool, optional
            If True, then robust SE are computed instead. Robustness here
            means that MLR-esque sandwich correction is applied. The default
            is False.

        Returns
        -------
        pd.DataFrame | dict
            Dataframe or mapping matrix_name->matrix.
    
        """
        from .inspector import inspect
        return inspect(self, mode=mode, what=what, information=information,
                       std_est=std_est, se_robust=se_robust)

    def operation_start(self, operation):
        """
        Works through START command.

        Sets starting values to parameters.
        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Raise
        ----------
        IndexError
            When no starting value is supplied as an argument to START.
        KeyError
            When invalid parameter name is supplied.

        Returns
        -------
        None.

        """
        try:
            start = operation.params[0]
        except IndexError:
            raise IndexError('START must have starting value as an argument.')
        for param in operation.onto:
            try:
                self.parameters[param].start = start
            except KeyError:
                raise KeyError(f'{param} is not a valid parameter name.')

    def operation_bound(self, operation):
        """
        Works through BOUND command.

        Sets bound constraints to parameters.
        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Raise
        ----------
        SyntaxError
            When no starting value is supplied as an argument to START.
        ValueError
            When BOUND arguments are not translatable through FLOAT.
        KeyError
            When invalid parameter name is supplied.

        Returns
        -------
        None.

        """
        try:
            a = float(operation.params[0])
            b = float(operation.params[1])
            b = (a, b)
        except IndexError:
            raise SyntaxError('BOUND must have 2 bounding arguments.')
        except ValueError:
            raise ValueError('BOUND arguments must be floats.')
        for param in operation.onto:
            try:
                self.parameters[param].bound = b
            except KeyError:
                raise KeyError(f'{param} is not a valid parameter name.')

    def operation_constraint(self, operation):
        """
        Works through CONSTRAINT command.

        Adds inequality and equality constraints.
        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Raise
        ----------
        SyntaxError
            When no starting value is supplied as an argument to START.

        Returns
        -------
        None.

        """
        try:
            constr = operation.params[0]
        except IndexError:
            raise SyntaxError('CONSTRAINT must have 1 argument: constraint.')
        params = [name for name, param in self.parameters.items()
                  if param.active]
        self.constraints.append(parse_constraint(constr, params))

    def operation_define(self, operation):
        """
        Works through DEFINE command.

        Here, used to add ordinal variables to the variable holder.
        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Returns
        -------
        None.

        """
        if operation.params and operation.params[0] == 'ordinal':
            if 'ordinal' not in self.vars:
                ords = set()
                self.vars['ordinal'] = ords
            else:
                ords = self.vars['ordinal']
            ords.update(operation.onto)  

    def update_matrices(self, params: np.ndarray):
        """
        Update all matrices from a parameter vector.

        Parameters
        ----------
        params : np.ndarray
            Vector of parameters.

        Returns
        -------
        None.

        """
        for mx, (r1, r2) in zip(self.matrices, self.param_ranges):
            if r1:
                mx[r1] = params[r2]

    def load_starting_values(self):
        """
        Load starting values for parameters from empirical data.

        Returns
        -------
        None.

        """
        for _, param in self.parameters.items():
            if param.start is None:
                loc = param.locations[0]
                n = next(n for n in range(len(self.matrices))
                         if self.matrices[n] is loc.matrix)
                row, col = self.names[n]
                lval, rval = row[loc.indices[0]], col[loc.indices[1]]
                param.start = self.start_rules[n](self, lval, rval)
            for loc in param.locations:
                loc.matrix[loc.indices] = param.start
                if loc.symmetric:
                    loc.matrix[loc.indices[::-1]] = param.start

    def load_data(self, data: pd.DataFrame, covariance=None, groups=None):
        """
        Load dataset from data matrix.

        Parameters
        ----------
        data : pd.DataFrame
            Dataset with columns as variables and rows as observations.
        covariance : pd.DataFrame, optional
            Custom covariance matrix. The default is None.
        groups : list, optional
            List of group names to center across. The default is None.

        Returns
        -------
        None.

        """
        if groups is None:
            groups = list()
        obs = self.vars['observed']
        for group in groups:
            for g in data[group].unique():
                inds = data[group] == g
                if sum(inds) == 1:
                    continue
                data.loc[inds, obs] -= data.loc[inds, obs].mean()
                data.loc[inds, group] = g
        self.mx_data = data[obs].values
        if len(self.mx_data.shape) != 2:
            self.mx_data = self.mx_data[:, np.newaxis]
        if 'ordinal' not in self.vars:
            self.load_cov(covariance.loc[obs, obs].values
                          if covariance is not None else cov(self.mx_data))
        else:
            inds = [obs.index(v) for v in self.vars['ordinal']]
            self.load_cov(hetcor(self.mx_data, inds))
        self.n_samples, self.n_obs = self.mx_data.shape

    def load_dataset(self, data: pd.DataFrame, ordcor=None, **kwargs):
        """
        Load dataset.

        Backward-compatibility method for semopy 1.+.
        Parameters
        ----------
        data : pd.DataFrame
            Dataset.
        ordcor : bool, optional
            If iterable, then it lists a set of ordinal variables. The default
            is None.

        Returns
        -------
        None.

        """
        if ordcor:
            try:
                vs = self.vars['ordinal']
            except KeyError:
                vs = set()
                self.vars['ordinal'] = vs
            for var in ordcor:
                vs.add(var)
        self.load(data)

    def load_cov(self, covariance: np.ndarray):
        """
        Load covariance matrix.

        Parameters
        ----------
        covariance : np.ndarray
            Covariance matrix.

        Returns
        -------
        None.

        """
        if type(covariance) is pd.DataFrame:
            obs = self.vars['observed']
            covariance = covariance.loc[obs, obs].values
        if covariance.size == 1:
            covariance.resize((1, 1))
        self.mx_cov = covariance
        try:
            self.mx_cov_inv, self.cov_logdet = chol_inv2(self.mx_cov)
        except np.linalg.LinAlgError:
            logging.warning('Sample covariance matrix is not PD. It may '
                            'indicate that data is bad. Also, it arises often '
                            'when polychoric/polyserial correlations are used.'
                            ' semopy now will run nearPD subroutines.')
            self.mx_cov = cov_nearest(covariance, threshold=1e-2)
            self.mx_cov_inv, self.cov_logdet = chol_inv2(self.mx_cov)
        self.mx_covlike_identity = np.identity(self.mx_cov.shape[0])

    def get_bounds(self):
        """
        Get bound constraints on parameters.

        Returns
        -------
        list
            List of tuples specifying bounds.

        """
        return [param.bound for _, param in self.parameters.items()
                if param.active]

    def load(self, data=None, cov=None, groups=None, clean_slate=False,
             n_samples=None):
        """
        Load dataset.

        Either data or cov must be provided. If both are provided, covariance
        matrix is not estimated from data, but data is used for starting values
        initialization. If only cov is provided, it's impossible to estimate
        starting values for regression coefficients, also attempts to estimate
        polychoric/polyserial correlations if requested are ignored.
        Parameters
        ----------
        data : pd.DataFrame, optional
            Data with columns as variables. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. The default is None.
        groups : list, optional
            Groups of size > 1 to center across. The default is None.
        clean_slate : bool, optional
            If True, resets parameters vector. The default is False.
        n_samples : int, optional
            Number of samples in data. Used only if data is None and cov is
            provided for Fisher Information Matrix calculation. The default is
            None.

        Raises
        ------
        Exception
            Rises when both data and cov are not provided or when FIML is used
            in abscence of full data.

        KeyError
            Rises when there are missing variables from the data.

        Returns
        -------
        None.

        """
        if data is None and cov is None:
            has_data = hasattr(self, 'mx_data')
            has_cov = hasattr(self, 'mx_cov')
            if not has_data and not has_cov:
                raise Exception('Either data or cov must be provided.')
            if clean_slate:
                self.prepare_params()
            return
        if data is None:
            logging.info('Providing only covariance matrix is unadvised as it\
                         prevents from estimating good starting values for\
                         loadings.')
        else:
            data = data.copy()
        cols = data.columns if data is not None else cov.columns
        obs = self.vars['observed']
        missing = set(obs) - set(set(cols))
        if missing:
            t = ', '.join(missing)
            raise KeyError('Variables {} are missing from data.'.format(t))
        if data is not None:
            self.load_data(data, covariance=cov, groups=groups)
        else:
            self.load_cov(cov)
            self.n_samples = n_samples
            if groups is not None:
                logging.warning('"groups" argument is redunant with cov \
                                matrix.')
        if (data is not None) or (cov is not None):
            self.load_starting_values()
        if (clean_slate) or (not hasattr(self, 'param_vals')):
            self.prepare_params()

    def fit(self, data=None, cov=None, obj='MLW', solver='SLSQP', groups=None,
            clean_slate=False, regularization=None, n_samples=None, **kwargs):
        """
        Fit model to data.

        Parameters
        ----------
        data : pd.DataFrame, optional
            Data with columns as variables. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. The default is None.
        obj : str, optional
            Objective function to minimize. Possible values are 'MLW', 'FIML',
            'ULS', 'GLS', 'WLS', 'DWLS'. The default is 'MLW'.
        solver : str, optional
            Optimizaiton method. Currently scipy-only methods are available.
            The default is 'SLSQP'.
        groups : list, optional
            Groups of size > 1 to center across. The default is None.
        clean_slate : bool, optional
            If False, successive fits will be performed with previous results
            as starting values. If True, parameter vector is reset each time
            prior to optimization. The default is False.
        regularization
            Special structure as returend by create_regularization function.
            If not None, then a regularization will be applied to a certain
            parameters in the model. The default is None.
        n_samples : int, optional
            Number of samples in data. Used only if data is None and cov is
            provided for Fisher Information Matrix calculation. The default is
            None.

        Raises
        ------
        Exception
            Rises when attempting to use FIML in absence of full data.

        Returns
        -------
        SolverResult
            Information on optimization process.

        """
        self.load(data=data, cov=cov, groups=groups,
                  clean_slate=clean_slate, n_samples=n_samples)
        if obj == 'FIML':
            if not hasattr(self, 'mx_data'):
                raise Exception('Full data must be supplied for FIML')
            self.prepare_fiml()
        elif obj in ('WLS', 'DWLS'):
            if (not hasattr(self, 'last_result')) or \
                (self.last_result.name_obj != obj):
                    self.prepare_wls(obj, data is None)
        fun, grad = self.get_objective(obj, regularization=regularization)
        solver = Solver(solver, fun, grad, self.param_vals,
                        constrs=self.constraints,
                        bounds=self.get_bounds(), **kwargs)
        res = solver.solve()
        res.name_obj = obj
        self.param_vals = res.x
        self.update_matrices(res.x)
        self.last_result = res
        return res

    def get_objective(self, name: str, regularization=None):
        """
        Retrieve objective function and its gradient by name.

        Parameters
        ----------
        name : str
            Name of objective function.
        regularization
            Special structure as returend by create_regularization function.
            If not None, then a regularization will be applied to a certain
            parameters in the model. The default is None.

        Raises
        ------
        KeyError
            Rises if incorrect name is provided.

        Returns
        -------
        tuple
            Objective function and gradient function.

        """
        try:
            if regularization is None:
                return self.objectives[name]
            else:
                fun, grad = self.objectives[name]
                regu, regu_grad = regularization
                if regu_grad is None:
                    return lambda x: fun(x) + regu(x), None
                else:
                    return lambda x: fun(x) + regu(x),\
                        lambda x: grad(x) + regu_grad(x)
        except KeyError:
            raise KeyError(f'{name} is unknown objective function.')

    def prepare_fiml(self):
        """
        Prepare data structure for efficient FIML estimation.

        Returns
        -------
        None.

        """
        d = defaultdict(list)
        for i in range(self.mx_data.shape[0]):
            t = tuple(list(np.where(np.isfinite(self.mx_data[i]))[0]))
            d[t].append(i)
        for cols, rows in d.items():
            inds = tuple(i for i in range(self.mx_data.shape[1])
                         if i not in cols)
            t = self.mx_data[rows][:, cols]
            d[cols] = (t.T @ t, inds, len(rows))
        self.fiml_data = d

    def prepare_wls(self, obj: str, custom_w=None):
        """
        Prepare data structures for efficient WLS/DWLS estimation.

        Parameters
        ----------
        obj : str
            Either 'WLS' or 'DWLS'.
        custom_w : np.ndarray, optional
            Optional custom weight matrix. The default is None.

        Returns
        -------
        None.

        """
        
        data = self.mx_data - self.mx_data.mean(axis=0)
        products = list()
        for i in range(data.shape[1]):
            for j in range(i, data.shape[1]):
                products.append(data[:, i] * data[:, j])
        products = np.array(products)
        w = np.cov(products, bias=True)
        if obj == 'DWLS':
            self.mx_w_inv = np.array([1 / d for d in w.diagonal()])
        else:
            try:
                self.mx_w_inv = np.linalg.inv(w)
            except np.linalg.LinAlgError:
                logging.warning("Weight matrix could not be inverted. NearPD"
                                " estimate will be used instead.")
                w = cov_nearest(w, threshold=1e-2)
                self.mx_w_inv = np.linalg.pinv(w)
        self.mx_w = w
        self.inds_triu_sigma = np.triu_indices_from(self.mx_cov)
        self.mx_vech_s = self.mx_cov[self.inds_triu_sigma]


    def predict(self, x: pd.DataFrame, intercepts=False):
        """
        Predict data given certain observations.
        
        Uses conditional expectation of the normal distribution method.

        Parameters
        ----------
        x : pd.DataFrame
            DataFrame with missing variables either not present at all, or
            with missing entries set to NaN.
        intercepts : bool, optional
            If True, then intercepts will be estimated and used in prediction
            procedure.

        Returns
        -------
        None.

        """
        sigma, (m, c) = self.calc_sigma()
        obs = self.vars['observed']
        result = x.copy()
        for v in obs:
            if v not in result:
                result[v] = np.nan
        result = result[obs].values
        d = dict()
        if intercepts:
            from .means import estimate_means
            means = estimate_means(self)
        for i in range(len(result)):
            pr = np.isfinite(result[i])
            present = tuple(pr)
            if present in d:
                continue
            if np.all(pr):
                d[present] = (None, None, None)
                continue
            missing = ~pr
            sigma12 = sigma[missing][:, pr]
            sigma22 = np.linalg.pinv(sigma[pr][:, pr])
            d[present] = (pr, missing, sigma12 @ sigma22)
        for i in range(len(result)):
            row = result[i]
            present = tuple(np.isfinite(row))
            present, missing, sprod = d[tuple(present)]
            if present is None:
                continue
            if intercepts:
                lvals = list(means['lval'])
                ests = list(means['Estimate'])
                ins = list()
                ob = list()
                for v in self.vars['inner']:
                    try:
                        i = lvals.index(v)
                        ins.append(ests[i])
                    except ValueError:
                        ins.append(0)
                for v in obs:
                    if v in self.vars['_output']:
                        try:
                            i = lvals.index(v)
                            ob.append(ests[i])
                        except ValueError:
                            ob.append(0)
                    else:
                        ob.append(0)
                mn = m @ np.array(ins) + np.array(ob)
                c = row[present] - mn[present]
                row[missing] = mn[missing] + sprod @ c
            else:
                row[missing] = sprod @ row[present]
        return pd.DataFrame(result, columns=obs)

    def predict_general(self, x: pd.DataFrame, solver='SLSQP', factors=True,
                        ret_opt=False, chunk_size=20):
        """
        Impute/predict data given certain observations. Very general method,
        and quiet slow and prone to errors, too.

        With Model, it might  be better to center x beforehand for factors
        to have zero mean.
        Warning: if you seek to compute only factor scores, predict_factors
        is much more preferable as it is way more faster and stable. 
        Parameters
        ----------
        x : pd.DataFrame
            Observations with possibly missing entries of certain variables.
            It's possible to provide latent variables "observations" too.
        solver : str, optional
            Solver to use. The default is 'SLSQP'.
        factors: bool
            If True, factor scores are estimated. The default is True.
        ret_opt : SolverResult, optional
            If True, SolverResult is also returned. The default is False.
        chunk_size : int, optional
            Changes the number of individuals to be processed at once.
            In theory, should only affect performance. If None, then the whole
            dataset is procssed at once. The default is 40.
        Returns
        -------
        pd.DataFrame
            Table with imputed data.

        """
        if chunk_size is not None:
            data = None
            opts = list()
            for i in range(0, x.shape[0], chunk_size):
                res = self.predict(x.iloc[i:i + chunk_size], solver=solver,
                                   factors=factors, ret_opt=ret_opt, 
                                   chunk_size=None)
                if ret_opt:
                    opts.append(res[1])
                    res = res[0]
                if data is None:
                    data = res
                else:
                    data = data.append(res)
            return (data, opts) if ret_opt else data
        from .imputer import get_imputer
        imp = get_imputer(self)(self, x, factors=factors)
        res = imp.fit(solver='SLSQP')
        data = imp.get_fancy()
        return data if not ret_opt else (data, res)

    def predict_factors(self, x: pd.DataFrame, center=True):
        """
        Fast factor estimation method via MAP. Requires complete data.

        Parameters
        ----------
        x : pd.DataFrame
            Complete data of observed variables.
        center : bool, optional
            If True, factors are centered before being returned. It might be
            helpful if x is not centered, whereas Model does not work with
            intercepts.

        Returns
        -------
        Factor scores.

        """
        lats = self.vars['latent']
        num_lat = len(lats)
        if num_lat == 0:
            return pd.DataFrame([])
        inners = self.vars['inner']
        obs = self.vars['observed']
        x = x[obs].values.T
        lambda_h = self.mx_lambda[:, :num_lat]
        lambda_x = self.mx_lambda[:, num_lat:]
        c = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        c_1 = c[:num_lat, :]
        c_2 = c[num_lat:, :]
        M_h = x
        t = lambda_x @ c_2
        L_zh = (t @ self.mx_psi @ t.T + self.mx_theta) * (x.shape[1])
        tr_lzh = np.trace(L_zh)
        tr_sigma = tr_lzh / x.shape[1]
        L_zh = np.linalg.inv(L_zh)
        t = lambda_h.T @ L_zh
        A = (tr_lzh / tr_sigma) * t @ M_h
        A_0 = tr_lzh * t @ lambda_h
        L_h = np.linalg.inv(c_1 @ self.mx_psi @ c_1.T)
        H = np.linalg.inv(A_0 / tr_sigma + L_h) @ A
        H = pd.DataFrame(H.T, columns=filter(lambda v: v in lats, inners))
        if center:
            H -= H.mean()
        return H

    '''
    ----------------------------LINEAR ALGEBRA PART---------------------------
    ----------------------The code below is responsible-----------------------
    ------------------for covariance structure computations-------------------
    '''

    def calc_sigma(self):
        """
        Calculate model-implied covariance matrix.

        Returns
        -------
        sigma : np.ndarray
            Sigma model-implied covariance matrix.
        tuple
            Tuple of auxiliary matrics Lambda @ C and C, where C = (I - B)^-1.

        """
        beta, lamb, psi, theta = self.matrices[:4]
        c = np.linalg.inv(self.identity_c - beta)
        m = lamb @ c
        return m @ psi @ m.T + theta, (m, c)

    def calc_sigma_grad(self, m: np.ndarray, c: np.ndarray):
        """
        Calculate gradient wrt to parameters vector of Sigma matrix.

        Parameters
        ----------
        m : np.ndarray
            Auxilary matrix returned from calc_sigma Lambda @ C.
        c : np.ndarray
            Auxilary matrix returned from calc_sigma (I - B)^-1

        Returns
        -------
        grad : list
            List of derivatives of Sigma wrt to parameters vector.

        """
        psi = self.matrices[2]
        m_t = m.T
        p = c @ psi
        d = p @ m_t
        grad = list()
        for dmxs in self.mx_diffs:
            g = np.float32(0.0)
            if dmxs[0] is not None:  # Beta
                t = dmxs[0] @ p
                g += m @ (t + t.T) @ m_t
            if dmxs[1] is not None:  # Lambda
                t = dmxs[1] @ d
                g += t + t.T
            if dmxs[2] is not None:  # Psi
                g += m @ dmxs[2] @ m_t
            if dmxs[3] is not None:  # Theta
                g += dmxs[3]
            grad.append(g)
        return grad

    '''
    ----------------------The code below is responsible-----------------------
    ------------------for objective function computations---------------------
    '''

    '''
    ------------------------Wishart logratio function-------------------------
    '''

    def obj_mlw(self, x: np.ndarray):
        """
        Calculate Wishart likelihood logratio.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Logratio of likelihoods.

        """
        self.update_matrices(x)
        try:
            sigma, _ = self.calc_sigma()
            inv_sigma, logdet_sigma = chol_inv2(sigma)
        except np.linalg.LinAlgError:
            return np.inf
        log_det_ratio = logdet_sigma - self.cov_logdet
        tr = np.einsum('ij,ji->', self.mx_cov, inv_sigma) - sigma.shape[0]
        loss = tr + log_det_ratio
        if loss < 0:
            return np.inf
        return loss

    def grad_mlw(self, x: np.ndarray):
        """
        Calculate Wishart likelihood logratio gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of MLW.

        """
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
            inv_sigma = chol_inv(sigma)
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        sigma_grad = self.calc_sigma_grad(m, c)
        cs = inv_sigma - inv_sigma @ self.mx_cov @ inv_sigma
        return np.array([np.einsum('ij,ji->', cs, g)
                         for g in sigma_grad])

    '''
    ---------------------Full Information Maximum Likelihood-------------------
    '''

    def obj_fiml(self, x: np.ndarray):
        """
        Calculate FIML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            FIML.

        """
        self.update_matrices(x)
        sigma_full, _ = self.calc_sigma()
        tr = 0
        logdet = 0
        for _, (mx, inds, n) in self.fiml_data.items():
            sigma = delete_mx(sigma_full, inds)
            try:
                sigma_inv, logdet_sigma = chol_inv2(sigma)
            except np.linalg.LinAlgError:
                return np.inf
            tr += np.einsum('ij,ji->', mx, sigma_inv)
            logdet += n * logdet_sigma
        loss = tr + logdet
        if loss < 0:  # Realistically should never happen.
            return np.inf
        return loss

    def grad_fiml(self, x: np.ndarray):
        """
        Calculate FIML gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of FIML.

        """
        self.update_matrices(x)
        sigma_full, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        grad = [0.0] * len(sigma_grad)
        for _, (mx, inds, n) in self.fiml_data.items():
            sigma = delete_mx(sigma_full, inds)
            try:
                sigma_inv = chol_inv(sigma)
            except np.linalg.LinAlgError:
                t = np.zeros(len(grad))
                t[:] = np.inf
                return t
            t = sigma_inv @ mx @ sigma_inv
            for i, g in enumerate(sigma_grad):
                g = delete_mx(g, inds)
                tr = -np.einsum('ij,ji->', t, g)
                logdet = n * np.einsum('ij,ji->', sigma_inv, g)
                grad[i] += tr + logdet
        return np.array(grad)

    '''
    -------------------------Unweighted Least Squares-------------------------
    '''

    def obj_uls(self, x: np.ndarray):
        """
        Calculate ULS objective value.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            ULS value.

        """
        self.update_matrices(x)
        try:
            sigma, _ = self.calc_sigma()
        except np.linalg.LinAlgError:
            return np.inf
        t = sigma - self.mx_cov
        loss = np.einsum('ij,ij->', t, t)
        return loss

    def grad_uls(self, x: np.ndarray):
        """
        Gradient of ULS objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of ULS.

        """
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        sigma_grad = self.calc_sigma_grad(m, c)
        t = sigma - self.mx_cov
        return 2 * np.array([np.einsum('ij,ji->', g, t)
                             for g in sigma_grad])

    '''
    -------------------------Generalized Least Squares------------------------
    '''

    def obj_gls(self, x: np.ndarray):
        """
        Calculate GLS objective value.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            GLS value.

        """
        self.update_matrices(x)
        try:
            sigma, _ = self.calc_sigma()
        except np.linalg.LinAlgError:
            return np.inf
        t = sigma @ self.mx_cov_inv - self.mx_covlike_identity
        return np.einsum('ij,ji->', t, t)

    def grad_gls(self, x: np.ndarray):
        """
        Gradient of GLS objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of GLS.

        """
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        sigma_grad = self.calc_sigma_grad(m, c)
        t = self.mx_cov_inv @ \
            (sigma @ self.mx_cov_inv - self.mx_covlike_identity)
        return 2 * np.array([np.einsum('ij,ji->', g, t)
                             for g in sigma_grad])

    '''
    -------------------------Weighted Least Squares------------------------
    '''

    def obj_wls(self, x: np.ndarray):
        self.update_matrices(x)
        try:
            diff = self.calc_sigma()[0][self.inds_triu_sigma] - self.mx_vech_s
        except np.linalg.LinAlgError:
            return np.nan
        return diff.T @ self.mx_w_inv @ diff
    
    def grad_wls(self, x: np.ndarray):
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
            sigma = sigma[self.inds_triu_sigma]
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        sigma_grad = [g[self.inds_triu_sigma].T
                        for g in self.calc_sigma_grad(m, c)]
        diff = sigma - self.mx_vech_s
        t = self.mx_w_inv @ diff
        return 2 * np.array([g @ t for g in sigma_grad])
    
    def obj_dwls(self, x: np.ndarray):
        self.update_matrices(x)
        try:
            diff = self.calc_sigma()[0][self.inds_triu_sigma] - self.mx_vech_s
        except np.linalg.LinAlgError:
            return np.nan
        return diff.T * self.mx_w_inv @ diff
    
    def grad_dwls(self, x: np.ndarray):
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
            sigma = sigma[self.inds_triu_sigma]
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        sigma_grad = [g[self.inds_triu_sigma].T
                        for g in self.calc_sigma_grad(m, c)]
        diff = sigma - self.mx_vech_s
        t = self.mx_w_inv * diff
        return 2 * np.array([g @ t for g in sigma_grad])
        

    '''
    -------------------------Fisher Information Matrix------------------------
    '''

    def calc_fim(self, inverse=False):
        """
        Calculate Fisher Information Matrix.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        sigma, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        inv_sigma = chol_inv(sigma)
        sz = len(sigma_grad)
        info = np.zeros((sz, sz))
        sgs = [sg @ inv_sigma for sg in sigma_grad]
        n = self.n_samples
        if n is None:
            raise AttributeError('For FIM estimation in a covariance-matr'
                                 'ix-only setting, you must provide the'
                                 ' n_samples argument to the fit or load'
                                 ' methods.')
        n /= 2
        
        for i in range(sz):
            for k in range(i, sz):
                info[i, k] = n * np.einsum('ij,ji->', sgs[i], sgs[k])
        fim = info + np.triu(info, 1).T
        if inverse:
            try:
                fim_inv = chol_inv(fim)
                self._fim_warn = False
            except np.linalg.LinAlgError:
                logging.warn("Fisher Information Matrix is not PD."
                              "Moore-Penrose inverse will be used instead of "
                              "Cholesky decomposition. See "
                              "10.1109/TSP.2012.2208105.")
                self._fim_warn = True
                fim_inv = np.linalg.pinv(fim)
            return (fim, fim_inv)
        return fim

    def grad_se_g(self, x: np.ndarray):
        """
        Calculate a list of separate likelihoods for each observation.

        A helper function that might be used to estimate Huber-White sandwich
        corrections.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        list
            List of len n_samples.

        """
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
            sigma_grad = self.calc_sigma_grad(m, c)
            inv_sigma = np.linalg.pinv(sigma)
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        res = list()
        mx_i = np.identity(sigma.shape[0])
        data = self.mx_data.copy()
        data -= data.mean(axis=0)
        for i in range(self.mx_data.shape[0]):
            x = data[i, np.newaxis]
            t = inv_sigma @ (mx_i -  x.T @ x @ inv_sigma)
            res.append(np.array([np.einsum('ij,ji->', t, g)
                                 for g in sigma_grad]) / 2)      
        return res


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Augmented semopy model with mean structure."""
import logging
import numpy as np
import pandas as pd
from .model import Model
from collections import defaultdict
from scipy.linalg import block_diag
from .utils import chol, chol_inv, chol_inv2, delete_mx, cov


class ModelMeans(Model):
    """
    Model with a mean stucture.

    Augmented model with exogenous variables ruled out into a separate
    structure. This, in turn, results in all exogenous variables being de-facto
    absent from the model, which may lead to less computations in certain
    cases (i.e. when a number of observed exogenous variables is huge, for
    example in GWAS).
    """

    matrices_names = tuple(list(Model.matrices_names) + ['gamma1', 'gamma2'])

    def __init__(self, description: str, mimic_lavaan=False, baseline=False,
                 cov_diag=False, intercepts=True):
        """
        Instantiate Model with mean-structure.

        Parameters
        ----------
        description : str
            Model description in semopy syntax.
        mimic_lavaan: bool
            If True, output variables are correlated and not conceptually
            identical to indicators. lavaan treats them that way, but it's
            less computationally effective. The default is False.
        baseline : bool
            If True, the model will be set to baseline model.
            Baseline model here is an independence model where all variables
            are considered to be independent with zero covariance. Only
            variances are estimated. The default is False.
        cov_diag : bool
            If cov_diag is True, then there are no covariances parametrised
            unless explicitly specified. The default is False.
        intercepts: bool
            If True, intercepts are also modeled. Intercept terms can be
            accessed via "1" symbol in a regression equation, i.e. x1 ~ 1. The
            default is True.

        Returns
        -------
        None.

        """
        self.intercepts = intercepts
        self.calc_fim = self.calc_fim_ml
        super().__init__(description, mimic_lavaan=mimic_lavaan,
                         cov_diag=cov_diag, baseline=baseline)
        self.objectives = {'FIML': (self.obj_fiml, self.grad_fiml),
                           'REML': (self.obj_reml, self.grad_reml),
                           'GLS': (self.obj_gls, self.grad_gls)}


    def preprocess_effects(self, effects: dict):
        """
        Run a routine just before effects are applied.

        Used to apply covariances to model.
        Parameters
        -------
        effects : dict
            Mapping opcode->lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        super().preprocess_effects(effects)
        if self.intercepts:
            for v in self.vars['observed']:
                if v not in self.vars['latent']:  # Workaround for Imputer
                    t = effects[self.symb_regression][v]
                    if '1' not in t:
                        t['1'] = None
        gamma1 = set()  # Those that load onto inner variables
        gamma2 = set()  # Those that load onto output variables
        for lval, rvs in effects[self.symb_regression].items():
            rvals = filter(lambda v: v in self.vars['observed_exogenous'], rvs)
            if lval in self.vars['_output']:
                gamma2.update(rvals)
            else:
                gamma1.update(rvals)
        self.vars['observed_exogenous'] = sorted(gamma1 | gamma2)
        self.vars['observed_exogenous_1'] = gamma1
        self.vars['observed_exogenous_2'] = gamma2

    def finalize_variable_classification(self, effects: dict):
        """
        Finalize variable classification.

        Reorders variables for better visual fancyness and does extra
        model-specific variable respecification.
        
        Parameters
        -------
        effects : dict
            Maping opcode->values->rvalues->mutiplicator.

        Returns
        -------
        None.

        """
        obs = self.vars['observed']
        exo = self.vars['exogenous']
        if self.intercepts:
            exo.add('1')
            obs.add('1')
        obs_exo = obs & exo
        to_rem = set()
        covs = effects[self.symb_covariance]
        for lv, rvs in covs.items():
            to_rem.add(lv)
            to_rem.update(rvs)
        obs_exo -= to_rem
        obs -= obs_exo
        exo -= obs_exo
        self.vars['observed_exogenous'] = obs_exo
        super().finalize_variable_classification(effects)

    def build_gamma1(self):
        """
        Gamma1 matrix contains relationships with exogenous variables.

        This Gamma1 matrix loads onto INNER variables, i.e. non-output
        variables.
        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        rows, cols = self.vars['inner'], self.vars['observed_exogenous']
        n, m = len(rows), len(cols)
        mx = np.zeros((n, m))
        return mx, (rows, cols)

    def build_gamma2(self):
        """
        Gamma2 matrix contains relationships with exogenous variables.

        This Gamma2 matrix loads onto OUTPUT variables.
        Returns
        -------
        np.ndarray
            Matrix.
        tuple
            Tuple of rownames and colnames.

        """
        rows, cols = self.vars['observed'], self.vars['observed_exogenous']
        n, m = len(rows), len(cols)
        mx = np.zeros((n, m))
        return mx, (rows, cols)

    def prepare_fiml(self):
        """
        Prepare data structure for efficient FIML calculation.

        Returns
        -------
        None.

        """
        d = defaultdict(list)
        data = self.mx_data
        for i in range(data.shape[0]):
            t = tuple(list(np.where(np.isfinite(data[i]))[0]))
            d[t].append(i)
        for cols, rows in d.items():
            inds = tuple(i for i in range(data.shape[1])
                         if i not in cols)
            t = data[rows, :][:, cols]
            d[cols] = (t.T, list(inds), rows)
        self.fiml_data = d

    def effect_regression(self, items: dict):
        """
        Work through regression operation.

        Parameters
        ----------
        items : dict
            Mapping lvalues->rvalues->multiplicator.

        Returns
        -------
        None.

        """
        items_super = defaultdict(dict)
        exo_obs = self.vars['observed_exogenous']
        inner = self.vars['inner']
        baseline = self.baseline
        for lv, rvs in items.items():
            for rv, mult in rvs.items():
                if baseline and rv != '1':
                    continue
                if rv not in exo_obs:
                    items_super[lv][rv] = mult
                    continue
                if lv in inner:
                    rows, cols = self.names_gamma1
                    i = rows.index(lv)
                    j = cols.index(rv)
                    mx = self.mx_gamma1
                else:
                    rows, cols = self.names_gamma2
                    i = rows.index(lv)
                    j = cols.index(rv)
                    mx = self.mx_gamma2
                ind = (i, j)
                name = None
                active = True
                try:
                    val = float(mult)
                    active = False
                except (TypeError, ValueError):
                    if mult is not None:
                        if mult == self.symb_starting_values:
                            active = False
                        else:
                            name = mult
                    val = None
                if name is None:
                    self.n_param_reg += 1
                    name = '_b%s' % self.n_param_reg
                self.add_param(name=name, matrix=mx, indices=ind, start=val,
                               active=active, symmetric=False,
                               bound=(None, None))
        super().effect_regression(items_super)

    def operation_define(self, operation):
        """
        Works through DEFINE command.

        Here, used to prevent user from attempting to use ordinal variables.
        Parameters
        ----------
        operation : Operation
            Operation namedtuple.

        Returns
        -------
        None.

        """
        if operation.params and operation.params[0] == 'ordinal':
            raise SyntaxWarning("Models with mean component do not support \
                                ordinal variables.")
        super().operation_define(operation)

    def load_data(self, data: pd.DataFrame, covariance=None, groups=None):
        """
        Load dataset from data matrix.

        Parameters
        ----------
        data : pd.DataFrame
            Dataset with columns as variables and rows as observations.
        covariance : pd.DataFrame, optional
            Custom covariance matrix. The default is None.
        groups : list, optional
            List of group names to center across. The default is None.

        Returns
        -------
        None.

        """
        if groups is None:
            groups = list()
        obs = self.vars['observed']
        for group in groups:
            for g in data[group].unique():
                inds = data[group] == g
                if sum(inds) == 1:
                    continue
                data.loc[inds, obs] -= data.loc[inds, obs].mean()
                data.loc[inds, group] = g
        self.mx_data = data[obs].values
        if len(self.mx_data.shape) != 2:
            self.mx_data = self.mx_data[:, np.newaxis]
        self.n_samples, self.n_obs = self.mx_data.shape
        self.mx_g = data[self.vars['observed_exogenous']].values.T
        if len(self.mx_g.shape) != 2:
            self.mx_g = self.mx_g[np.newaxis, :]
        g = self.mx_g
        if self.calc_fim == self.calc_fim_reml:
            try:
                s = np.identity(g.shape[1]) - g.T @ chol_inv(g @ g.T) @ g
            except ValueError:
                s = np.identity(g.shape[1]) - g.T @ g
            d, q = np.linalg.eigh(s)
            rank_dec = 0
            for i in d:
                if abs(i) < 1e-8:
                    rank_dec += 1
                else:
                    break
            d = np.diag(d)[rank_dec:, :]
            self.mx_s = d @ q.T
            self.mx_data_transformed = self.mx_s @ self.mx_data
            self.mx_data_square = self.mx_data_transformed.T @\
                                  self.mx_data_transformed
        self.load_cov(covariance.loc[obs, obs]
                      if covariance is not None else cov(self.mx_data))

    def load(self, data, cov=None, groups=None, clean_slate=False,
             n_samples=None):
        """
        Load dataset.

        Parameters
        ----------
        data : pd.DataFrame
            Data with columns as variables.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. Used only for variance
            starting values. The default is None.
        groups : list, optional
            Groups of size > 1 to center across. The default is None.
        clean_slate : bool, optional
            If True, resets parameters vector. The default is False.
        n_samples : int, optional
            Redunant for ModelMeans. The default is None.

        KeyError
            Rises when there are missing variables from the data.

        Returns
        -------
        None.

        """
        if data is None:
            if not hasattr(self, 'mx_data'):
                raise Exception("Data must be provided.")
            if clean_slate:
                self.prepare_params()
            return
        else:
            data = data.copy()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        if self.intercepts:
            data['1'] = 1.0
        cols = data.columns
        missing = (set(obs) | set(exo)) - set(cols)
        if missing:
            t = ', '.join(missing)
            raise KeyError('Variables {} are missing from data.'.format(t))
        self.load_data(data, covariance=cov, groups=groups)
        self.load_starting_values()
        if clean_slate or not hasattr(self, 'param_vals'):
            self.prepare_params()
        # Happens only if we ran fit with ML
        if self.calc_fim is self.calc_fim_ml:
            self.prepare_fiml()

    def fit(self, data=None, cov=None, obj='ML', solver='SLSQP', groups=None,
            clean_slate=False, **kwargs):
        """
        Fit model to data.

        Parameters
        ----------
        data : pd.DataFrame, optional
            Data with columns as variables. The default is None.
        cov : pd.DataFrame, optional
            Pre-computed covariance/correlation matrix. The default is None.
        obj : str, optional
            Objective function to minimize. Possible values are 'REML', "ML".
            The default is 'ML'.
        solver : TYPE, optional
            Optimizaiton method. Currently scipy-only methods are available.
            The default is 'SLSQP'.
        groups : list, optional
            Groups of size > 1 to center across. The default is None.
        clean_slate : bool, optional
            If False, successive fits will be performed with previous results
            as starting values. If True, parameter vector is reset each time
            prior to optimization. The default is False.

        Raises
        ------
        Exception
            Rises when attempting to use FIML in absence of full data.
        NotImplementedError
            Rises when unknown objective name is passed.

        Returns
        -------
        SolverResult
            Information on optimization process.

        """
        if obj == 'REML':
            self.calc_fim = self.calc_fim_reml
            res = super().fit(data=data, cov=cov, obj='REML', solver=solver,
                              groups=groups, clean_slate=clean_slate, **kwargs)
            sigma, (self.mx_m, self.mx_c) = self.calc_sigma()
            self.mx_sigma_inv = chol_inv(sigma)
            res_m = super().fit(obj='GLS', solver=solver,
                                groups=groups, clean_slate=False, **kwargs)
            return res, res_m
        elif obj == 'ML':
            self.calc_fim = self.calc_fim_ml
            res = super().fit(data=data, cov=cov, obj='FIML', solver=solver,
                              groups=groups, clean_slate=clean_slate, **kwargs)
            return res
        else:
            raise NotImplementedError(f"Unknown method {obj}.")

    '''
    ----------------------------LINEAR ALGEBRA PART---------------------------
    ----------------------The code below is responsible-----------------------
    ---------------------for mean structure computations------===-------------
    '''

    def calc_mean(self, m: np.ndarray):
        """
        Calculate mean component.

        Parameters
        ----------
        m : np.ndarray
            Lambda @ C.

        Returns
        -------
        np.ndarray
            Model-implied mean component.

        """
        return (m @ self.mx_gamma1 + self.mx_gamma2) @ self.mx_g

    def calc_mean_grad(self, m: np.ndarray, c: np.ndarray):
        """
        Calculate mean component gradient.

        Parameters
        ----------
        m : np.ndarray
            Lambda @ C.
        c : np.ndarray
            (I-B)^{-1}.

        Returns
        -------
        grad : list
            Gradient values of model-implied mean component.

        """
        grad = list()
        gm1_g1 = self.mx_gamma1 @ self.mx_g
        c_gm1_g1 = c @ gm1_g1
        for dmx in self.mx_diffs:
            g = np.float32(0.0)
            if dmx[0] is not None:  # Beta
                g += m @ dmx[0] @ c_gm1_g1
            if dmx[1] is not None:  # Lambda
                g += dmx[1] @ c_gm1_g1
            if dmx[4] is not None:  # Gamma1
                g += m @ dmx[4] @ self.mx_g
            if dmx[5] is not None:  # Gamma2
                g += dmx[5] @ self.mx_g
            grad.append(g)
        return grad

    def calc_mean_grad_reml(self):
        """
        Calculate mean component gradient given Sigma.

        Returns
        -------
        grad : list
            Gradient values of model-implied mean component.

        """
        grad = list()
        m = self.mx_m
        for dmx in self.mx_diffs:
            g = np.float32(0.0)
            if dmx[4] is not None:  # Gamma1
                g += m @ dmx[4] @ self.mx_g
            if dmx[5] is not None:  # Gamma2
                g += dmx[5] @ self.mx_g
            grad.append(g)
        return grad

    '''
    -----------------------Restricted Maximum Likelihood-----------------------
    '''

    def obj_gls(self, x: np.ndarray):
        """
        GLS objective for fitting mean component given known Sigma.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            GLS obective value.

        """
        self.update_matrices(x)
        sigma_inv = self.mx_sigma_inv
        m = self.mx_m
        mean = self.calc_mean(m)
        mx = self.mx_data
        center = mx - mean.T
        return np.einsum('ij,ji->', center.T @ center, sigma_inv)

    def grad_gls(self, x: np.ndarray):
        """
        Calculate GLS gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of GLS.

        """
        self.update_matrices(x)
        sigma_inv = self.mx_sigma_inv
        m = self.mx_m
        mean = self.calc_mean(m)
        mx = self.mx_data
        center = mx - mean.T
        t = center @ sigma_inv
        grad = list()
        for g in self.calc_mean_grad_reml():
            if len(g.shape):
                grad.append(-2 * np.einsum('ij,ji', g, t))
            else:
                grad.append(0.0)
        return np.array(grad)

    def obj_reml(self, x: np.ndarray):
        """
        Calculate REML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            REML.

        """
        self.update_matrices(x)
        sigma, _ = self.calc_sigma()
        tr = 0
        logdet = 0
        try:
            sigma_inv, logdet_sigma = chol_inv2(sigma)
        except np.linalg.LinAlgError:
            return np.inf
        tr += np.einsum('ij,ji->', self.mx_data_square, sigma_inv)
        logdet += self.mx_data_transformed.shape[0] * logdet_sigma
        loss = tr + logdet
        if loss < 0:
            return np.inf
        return loss

    def grad_reml(self, x: np.ndarray):
        """
        Calculate REML gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of REML.

        """
        self.update_matrices(x)
        sigma, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        try:
            sigma_inv = chol_inv(sigma)
        except np.linalg.LinAlgError:
            return np.array([np.inf] * len(x))
        n = self.mx_data_transformed.shape[0]
        cs = n * sigma_inv - sigma_inv @ self.mx_data_square @ sigma_inv
        grad = list()
        for dx in sigma_grad:
            if len(dx.shape):
                grad.append(np.einsum('ij,ji->', cs, dx))
            else:
                grad.append(0.0)
        return np.array(grad)

    '''
    ---------------------Full Information Maximum Likelihood-------------------
    '''

    def obj_fiml(self, x: np.ndarray):
        """
        Calculate FIML objective function.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            FIML.

        """
        self.update_matrices(x)
        sigma_full, (m, _) = self.calc_sigma()
        mean_full = self.calc_mean(m)
        tr = 0
        logdet = 0
        for _, (mx, inds, rows) in self.fiml_data.items():
            center = mx - np.delete(mean_full, inds, axis=0)[:, rows]
            s = center @ center.T
            sigma = delete_mx(sigma_full, inds)
            try:
                sigma_inv, logdet_sigma = chol_inv2(sigma)
            except np.linalg.LinAlgError:
                return np.inf
            tr += np.einsum('ij,ji->', s, sigma_inv)
            logdet += len(rows) * logdet_sigma
        loss = tr + logdet
        if loss < 0:  # Realistically should never happen.
            return np.inf
        return loss

    def grad_fiml(self, x: np.ndarray):
        """
        Calculate FIML gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Gradient of FIML.

        """
        self.update_matrices(x)
        sigma_full, (m, c) = self.calc_sigma()
        mean_full = self.calc_mean(m)
        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad(m, c)
        grad = [0.0] * len(sigma_grad)
        for _, (mx, inds, rows) in self.fiml_data.items():
            sigma = delete_mx(sigma_full, inds)
            try:
                sigma_inv = chol_inv(sigma)
            except np.linalg.LinAlgError:
                t = np.zeros(len(grad))
                t[:] = np.inf
                return t
            center = mx - np.delete(mean_full, inds, axis=0)[:, rows]
            s = center @ center.T
            a = center.T @ sigma_inv
            t = sigma_inv @ s @ sigma_inv
            n = len(rows)
            for i, (s_g, m_g) in enumerate(zip(sigma_grad, mean_grad)):
                if len(m_g.shape):
                    m_g = np.delete(m_g, inds, axis=0)[:, rows]
                    mean_tr = -2 * np.einsum('ij,ji->', m_g, a)
                else:
                    mean_tr = 0.0
                if len(s_g.shape):
                    s_g = delete_mx(s_g, inds)
                    tr = -np.einsum('ij,ji->', t, s_g)
                    logdet_tr = n * np.einsum('ij,ji->', sigma_inv, s_g)
                else:
                    logdet_tr = 0.0
                    tr = 0.0
                grad[i] += tr + logdet_tr + mean_tr
        return np.array(grad)

    '''
    -------------------------Prediction method--------------------------------
    '''

    def predict(self, x: pd.DataFrame):
        """
        Predict data given certain observations.
        
        Uses conditional expectation of the normal distribution method.

        Parameters
        ----------
        x : pd.DataFrame
            DataFrame with missing variables either not present at all, or
            with missing entries set to NaN.

        Returns
        -------
        None.

        """
        sigma, (m, _) = self.calc_sigma()
        obs = self.vars['observed']
        exo = self.vars['observed_exogenous']
        result = x.copy()
        for v in obs:
            if v not in result:
                result[v] = np.nan
        for v in exo:
            if v not in result:
                if v == '1':
                    result[v] = 1
                else:
                    result[v] = 0
        result = result[obs + exo]
        old_gamma = self.mx_g.copy()
        for i, (_, row) in enumerate(result.iterrows()):
            row_exo = row.loc[exo]
            row_endo = row.loc[obs]
            present = [True if np.isfinite(row[r]) else False
                       for r in obs]
            present = np.array(present)
            missing = ~present
            sigma12 = sigma[missing][:, present]
            sigma22 = np.linalg.pinv(sigma[present][:, present])
            self.mx_g = row_exo.values.reshape((-1, 1))
            mean = self.calc_mean(m)
            mean_m = mean[missing, :]
            mean_p = mean[present, :]
            p = mean_m
            if len(present):
                r = row_endo.iloc[present].values.reshape((-1, 1))
                p += sigma12 @ sigma22 @ (r - mean_p)
            row.iloc[np.where(missing)] = p.flatten()
            result.iloc[i] = row
        self.mx_g = old_gamma
        if '1' in result:
            result = result.drop('1', axis=1)
        return result

    def predict_exo(self, exogenous: pd.DataFrame):
        """
        Predict output variables given a set of exogenous variables.

        This method works much faster than "predict", however it can't be
        utilised to impute missing data or to estimate factors as of now. It is
        especially useful for phenotype prediction via a set of known SNPs.
        Parameters
        ----------
        exogenous : pd.DataFrame
            Observations of exogenous variables. Missing variables or mssing
            data are converted to zeros.

        Returns
        -------
        pd.DataFrame
            DataFrame containing predictions of endogenous observed variables.

        """
        exos = self.vars['observed_exogenous']
        exogenous = exogenous.copy()
        for v in exos:
            if v not in exogenous.columns:
                if v != '1':
                    exogenous[v] = 0
                else:
                    exogenous[v] = 1
        g = exogenous[exos].values.T 
        t = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        t = (self.mx_lambda @ t @ self.mx_gamma1 + self.mx_gamma2) @ g
        return pd.DataFrame(t.T, columns=self.vars['observed'],
                            index=exogenous.index)

    def predict_factors(self, x: pd.DataFrame):
        """
        Fast factor estimation method via MAP. Requires complete data.

        Parameters
        ----------
        x : pd.DataFrame
            Complete data of observed variables.

        Returns
        -------
        Factor scores.

        """
        lats = self.vars['latent']
        num_lat = len(lats)
        if num_lat == 0:
            return pd.DataFrame([])
        inners = self.vars['inner']
        obs = self.vars['observed']
        x = x[obs].values.T
        lambda_h = self.mx_lambda[:, :num_lat]
        lambda_x = self.mx_lambda[:, num_lat:]
        c = np.linalg.inv(np.identity(self.mx_beta.shape[0]) - self.mx_beta)
        c_1 = c[:num_lat, :]
        c_2 = c[num_lat:, :]
        g1 = self.mx_gamma1; g2 = self.mx_gamma2; g = self.mx_g
        M_h = x - (g2 + lambda_x @ c_2 @ g1) @ g
        t = lambda_x @ c_2
        L_zh = (t @ self.mx_psi @ t.T + self.mx_theta) * (x.shape[1])
        tr_lzh = np.trace(L_zh)
        tr_sigma = tr_lzh / x.shape[1]
        L_zh = np.linalg.inv(L_zh)
        t = lambda_h.T @ L_zh
        A = (tr_lzh / tr_sigma) * t @ M_h
        A_0 = tr_lzh * t @ lambda_h
        L_h = np.linalg.inv(c_1 @ self.mx_psi @ c_1.T)
        M = c_1 @ g1 @ g
        A_1 = L_h @ M
        H = np.linalg.inv(A_0 / tr_sigma + L_h) @ (A_1 + A)
        return pd.DataFrame(H.T, columns=filter(lambda v: v in lats, inners))

    '''
    -------------------------Fisher Information Matrix------------------------
    '''

    def calc_fim_reml(self, inverse=False):
        """
        Calculate Fisher Information Matrix when estimation was performed via
        REML.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        if not hasattr('self', 'mx_sigma_inv'):
            sigma, (m, c) = self.calc_sigma()
            sigma_inv = chol_inv(sigma)
        else:
            sigma_inv, (m, c) = self.mx_sigma_inv, self.mx_m, self.mx_c
        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad_reml()
        n = self.mx_data.shape[0] / 2
        inds_mean = list()
        inds_sigma = list()
        sgs, mgs = list(), list()
        for i, g in enumerate(sigma_grad):
            if len(g.shape):
                sgs.append(g @ sigma_inv)
                inds_sigma.append(i)
        for i, g in enumerate(mean_grad):
            if len(g.shape):
                mgs.append(sigma_inv @ g)
                inds_mean.append(i)
        sz = len(sgs)
        mx_var = np.zeros((sz, sz))
        for i in range(sz):
            for j in range(i, sz):
                mx_var[i, j] = n * np.einsum('ij,ji->', sgs[i], sgs[j])
        mx_var = mx_var + np.triu(mx_var, 1).T
        sz = len(mgs)
        mx_fixed = np.zeros((sz, sz))
        for i in range(sz):
            for j in range(i, sz):
                mx_fixed[i, j] = np.einsum('ij,ij->', mean_grad[inds_mean[i]],
                                           mgs[j])
        mx_fixed = mx_fixed + np.triu(mx_fixed, 1).T
        inds_mean = np.array(inds_mean, dtype=int)
        inds_sigma = np.array(inds_sigma, dtype=int)
        inds = np.append(inds_mean, inds_sigma)
        fim = block_diag(mx_fixed, mx_var)
        fim = fim[:, inds][:, inds]
        if inverse:
            try:
                mx_var_inv = chol_inv(mx_var)
                mx_fixed_inv = chol_inv(mx_fixed)
                self._fim_warn = False
            except np.linalg.LinAlgError:
                logging.warn("Fisher Information Matrix is not PD."
                             "Moore-Penrose inverse will be used instead of "
                             "Cholesky decomposition. See "
                              "10.1109/TSP.2012.2208105.")
                self._fim_warn = True
                mx_var_inv = np.linalg.pinv(mx_var)
                mx_fixed_inv = np.linalg.pinv(mx_fixed)
            fim_inv = block_diag(mx_fixed_inv, mx_var_inv)
            fim_inv = fim_inv[inds, :][:, inds]
            return (fim, fim_inv)
        return fim

    def calc_fim_ml(self, inverse=False):
        """
        Calculate Fisher Information Matrix when estimation was performed via
        ML.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        sigma, (m, c) = self.calc_sigma()
        sigma_grad = self.calc_sigma_grad(m, c)
        mean_grad = self.calc_mean_grad(m, c)
        inv_sigma = chol_inv(sigma)
        sz = len(sigma_grad)
        n = self.mx_data.shape[0] / 2
        info = np.zeros((sz, sz))
        sgs = [sg @ inv_sigma if len(sg.shape) else None for sg in sigma_grad]
        mgs = [inv_sigma @ g if len(g.shape) else None for g in mean_grad]
        for i in range(sz):
            for k in range(i, sz):
                if sgs[i] is not None and sgs[k] is not None:
                    info[i, k] = n * np.einsum('ij,ji->', sgs[i], sgs[k])
                if mgs[i] is not None and len(mean_grad[k].shape):
                    info[i, k] += np.einsum('ij,ij->', mean_grad[i], mgs[k])
        fim = info + np.triu(info, 1).T
        if inverse:
            try:
                fim_inv = chol_inv(fim)
                self.fim_warn = False
            except np.linalg.LinAlgError:
                logging.warn("Fisher Information Matrix is not PD."
                             "Moore-Penrose inverse will be used instead of "
                             "Cholesky decomposition. See "
                              "10.1109/TSP.2012.2208105.")
                self._fim_warn = True
                fim_inv = np.linalg.pinv(fim)
            return (fim, fim_inv)
        return fim


    def grad_se_g(self, x: np.ndarray):
        """
        Calculate a list of separate likelihoods for each observation.

        A helper function that might be used to estimate Huber-White sandwich
        corrections.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        list
            List of len n_samples.

        """
        self.update_matrices(x)
        try:
            sigma, (m, c) = self.calc_sigma()
            sigma_grad = self.calc_sigma_grad(m, c)
            mean_grad = self.calc_mean_grad(m, c)
            mean = self.calc_mean(m).T
            inv_sigma = np.linalg.pinv(sigma)
        except np.linalg.LinAlgError:
            t = np.zeros((len(x),))
            t[:] = np.inf
            return t
        res = list()
        mx_i = np.identity(sigma.shape[0])
        data = self.mx_data.copy()
        if not self.intercepts:
            data -= data.mean(axis=0)
        for i in range(self.mx_data.shape[0]):
            x = self.mx_data[i] - mean[i]
            x = x[:, np.newaxis]
            t = inv_sigma @ (mx_i -  x @ x.T @ inv_sigma)
            t2 = (inv_sigma @ x).flatten()
            g = np.zeros_like(self.param_vals)
            for i, (s_g, m_g) in enumerate(zip(sigma_grad, mean_grad)):
                if len(m_g.shape):
                    m = np.dot(t2, m_g[:, i])
                else:
                    m = 0.0
                if len(s_g.shape):
                    cov = np.einsum('ij,ji->', t, s_g) / 2
                else:
                    cov = 0
                g[i] += m + cov
            res.append(g)      
        return res


# -*- coding: utf-8 -*-
"""The module is responsible for generating parameters given model
description."""
from .description import generate_desc
from semopy import ModelMeans
import pandas as pd
import random


def param_variance():
    return random.uniform(0.7, 1.3)

def param_correlation():
    return random.uniform(-0.3, 0.3)

def param_loading():
    return random.choice([1, -1]) * random.uniform(0.3, 1.5)




random.seed(123)

desc = generate_desc(3, 2, 3, 3, 0)

def generate_parameters(desc: str, intercepts=False,
                        sampler_var_psi=param_variance,
                        sampler_cor_psi=param_correlation,
                        sampler_var_theta=param_variance,
                        sampler_cor_theta=param_correlation,
                        sampler_reg_beta=param_loading,
                        sampler_reg_lambda=param_loading,
                        sampler_reg_gamma=param_loading):
    """
    Generate random parameters for a given model.

    Parameters
    ----------
    desc : str
        Description of semopy model.
    intercepts : bool, optional
        Should intercepts be included in the model? The default is False.
    sampler_var_psi : callable, optional
        Method that samples variance for Psi matrix. The default is
        param_variance.
    sampler_cor_psi : callable, optional
        Method that samples correlation for Psi matrix.. The default is
        param_correlation.
    sampler_var_theta : callable, optional
        Method that samples variance for Theta matrix. The default is
        param_variance.
    sampler_cor_theta : callable, optional
        Method that samples correlation for Theta matrix. The default is
        param_correlation.
    sampler_reg_beta : callable, optional
        Method that samples loading for Beta matrix. The default is
        param_loading.
    sampler_reg_lambda : callable, optional
        Method that samples loading for Lambda matrix. The default is
        param_loading.
    sampler_reg_gamma : callable, optional
        Method that samples loading for Gamma matrix. The default is
        param_loading.

    Returns
    -------
    Pandas DataFrame with parameters values, dict of matrices, filled with
    parameters.
    """
    d = {'lval': [], 'op': [], 'rval': [], 'Estimate': []}
    m = ModelMeans(desc, intercepts=intercepts)
    matrices = m.inspect('mx', what='params')
    
    psi_rows = matrices['Psi'].index
    psi = matrices['Psi'].values
    for i in range(psi.shape[0]):
        if type(psi[i, i]) is str:
            p = m.parameters[psi[i, i]]
            if p.active:
                psi[i, i] = sampler_var_psi()
                d['lval'].append(psi_rows[i])
                d['op'].append('~~')
                d['rval'].append(psi_rows[i])
                d['Estimate'].append(psi[i, i])
            else:
                psi[i, i] = p.start
    for i in range(psi.shape[0]):
        for j in range(psi.shape[1]):
            if i != j and type(psi[i, j]) is str:
                p = m.parameters[psi[i, j]]
                if p.active:
                    t = (psi[i, i] * psi[j, j]) ** (0.5)
                    psi[i, j] = sampler_cor_psi() * t
                    psi[j, i] = psi[i, j]
                    d['lval'].append(psi_rows[i])
                    d['op'].append('~~')
                    d['rval'].append(psi_rows[j])
                    d['Estimate'].append(psi[i, j])
                else:
                    psi[i, j] = p.start
                    psi[j, i] = p.start
    m.mx_psi = psi.astype('float64')
    
    theta_rows = matrices['Theta'].index
    theta = matrices['Theta'].values
    for i in range(theta.shape[0]):
        if type(theta[i, i]) is str:
            p = m.parameters[theta[i, i]]
            if p.active:
                theta[i, i] = sampler_var_theta()
                d['lval'].append(theta_rows[i])
                d['op'].append('~~')
                d['rval'].append(theta_rows[i])
                d['Estimate'].append(theta[i, i])
            else:
                theta[i, i] = p.start
    for i in range(theta.shape[0]):
        for j in range(theta.shape[1]):
            if i != j and type(theta[i, j]) is str:
                p = m.parameters[theta[i, j]]
                if p.active:
                    t = (theta[i, i] * theta[j, j]) ** (0.5)
                    theta[i, j] = sampler_cor_theta() * t
                    theta[j, i] = theta[i, j]
                    d['lval'].append(theta_rows[i])
                    d['op'].append('~~')
                    d['rval'].append(theta_rows[j])
                    d['Estimate'].append(theta[i, j])
                else:
                    theta[i, j] = p.start
                    theta[j, i] = p.start
    m.mx_theta = theta.astype('float64')
    
    beta_rows = matrices['Beta'].index
    beta_cols = matrices['Beta'].columns
    beta = matrices['Beta'].values
    for i in range(beta.shape[0]):
        for j in range(beta.shape[1]):
            if type(beta[i, j]) is str:
                p = m.parameters[beta[i, j]]
                if p.active:
                    beta[i, j] = sampler_reg_beta()
                    d['lval'].append(beta_rows[i])
                    d['op'].append('~')
                    d['rval'].append(beta_cols[j])
                    d['Estimate'].append(beta[i, j])
                else:
                    beta[i, j] = p.start
    m.mx_beta = beta.astype('float64')

    lamb_rows = matrices['Lambda'].index
    lamb_cols = matrices['Lambda'].columns
    lamb = matrices['Lambda'].values
    s = set()
    firsts = m.first_manifs
    for j in range(lamb.shape[1]):
        for i in range(lamb.shape[0]):
            if type(lamb[i, j]) is str:
                p = m.parameters[lamb[i, j]]
                if p.active:
                    if firsts.get(lamb_cols[j], None) == lamb_rows[i]:
                        lamb[i, j] = 1.0
                        s.add(lamb_cols[j])
                    else:
                        lamb[i, j] = sampler_reg_lambda()
                    d['lval'].append(lamb_rows[i])
                    d['op'].append('~')
                    d['rval'].append(lamb_cols[j])
                    d['Estimate'].append(lamb[i, j])
                else:
                    lamb[i, j] = p.start
    m.mx_lambda = lamb.astype('float64')

    gamma_rows = matrices['Gamma1'].index
    gamma_cols = matrices['Gamma1'].columns
    gamma = matrices['Gamma1'].values
    for i in range(gamma.shape[0]):
        for j in range(gamma.shape[1]):
            if type(gamma[i, j]) is str:
                p = m.parameters[gamma[i, j]]
                if p.active:
                    gamma[i, j] = sampler_reg_gamma()
                    d['lval'].append(gamma_rows[i])
                    d['op'].append('~')
                    d['rval'].append(gamma_cols[j])
                    d['Estimate'].append(gamma[i, j])
                else:
                    gamma[i, j] = p.start
    m.mx_gamma1 = gamma.astype('float64')

    gamma_rows = matrices['Gamma2'].index
    gamma_cols = matrices['Gamma2'].columns
    gamma = matrices['Gamma2'].values
    for i in range(gamma.shape[0]):
        for j in range(gamma.shape[1]):
            if type(gamma[i, j]) is str:
                p = m.parameters[gamma[i, j]]
                if p.active:
                    gamma[i, j] = sampler_reg_gamma()
                    d['lval'].append(gamma_rows[i])
                    d['op'].append('~')
                    d['rval'].append(gamma_cols[j])
                    d['Estimate'].append(gamma[i, j])
                else:
                    gamma[i, j] = p.start
    m.mx_gamma2 = gamma.astype('float64')

    return pd.DataFrame.from_dict(d), m


# -*- coding: utf-8 -*-
"""Effect with a fully parametrised K matrix"""
from ..utils import calc_zkz
from .effect_base import EffectBase
import pandas as pd
import numpy as np


class EffectFree(EffectBase):
    """
    Effect with a fully parametrised K matrix.
    
    Number of new parameters introduced is equal to g(g + 1) / 2, where g is
    an unique number of groups in data.
    """
    def __init__(self, columns: str, correlation=True, d_mode='diag'):
        """
        Instantiate EffectStatic.

        Parameters
        ----------
        columns : str
            Name of column that corresponds to individuals group id. Should
            match the appropriate row/column in the K dataframe.
        correlation : bool, optional
            If True, then K is assumed to be a correlation matrix, i.e.
            diagonal parameters are fixed to 1 and non-diagonal ones are
            constrained to (-1; 1). The default is True;
        d_mode : str
            Mode of D matrix. If "diag", then D has unique params on the
            diagonal. If "full", then D is fully parametrised. If
            "scale", then D is an identity matrix, multiplied by a single
            variance parameter (scalar). The default is "diag".

        Returns
        -------
        None.

        """
        
        super().__init__(columns, d_mode=d_mode)
        self.correlation = correlation

    def load(self, i, model, data, clean_start=True, **kwargs):
        """
        Called by model new dataset is loaded.
        
        Here, Effects are configured from the data. self.parameters must be
        initialised after invoking this method.
        Parameters
        ----------
        order : int
            Identificator of effect in model. It is just an order of the effect
            among other effects as specified by user.
        model : ModelGeneralizedEffects
            Instance of ModelGeneralizedEffects that calls this method.
        data : pd.DataFrame
            Dataset that is being loaded. Should contain self.columns.
        clean_start : bool, optional
            If True, then parameters are (re)initialized. The model will use
            the ones already present in self.parameters vector otherwise. The
            default is True.

        Returns
        -------
        None.

        """
        super().load(i, model, data, clean_start, **kwargs)
        c = data[self.columns[0]]
        if clean_start:
            self.groups = list()
            for gt in c.unique():
                if type(gt) is str:
                    self.groups.extend(gt.split(';'))
                else:
                    if not np.isfinite(gt):
                        continue
                    self.groups.append(gt)
            self.groups = sorted(set(self.groups))
            p = len(self.groups)
            params = list()
            if not self.correlation:
                params.extend([0.1] * p)
            params.extend([0] * ((p * (p - 1)) // 2))
            self.num_p = p
            self.triu = np.triu_indices(p, 1)
            self.tril = np.tril_indices(p, -1)
            self.parameters = np.array(params)
        self.mx_z = calc_zkz(c, None, p_names=self.groups,
                             return_z=True)
        self.build_derivatives()

    def build_derivatives(self):
        z = self.mx_z
        grad = list()
        p = self.num_p
        if not self.correlation:
            for i in range(p):
                k = np.zeros((p, p))
                k[i, i] = 1
                grad.append(z @ k @ z.T)
        for a, b in np.nditer(self.triu):
            k = np.zeros((p, p))
            k[a, b] = 1
            k[b, a] = 1
            grad.append(z @ k @ z.T)
        self.grad = grad

    def calc_k(self, model):
        params = self.parameters
        p = self.num_p
        if not self.correlation:
            k = np.diag(params[:p])
            k[self.triu] = params[p:]
            k[self.tril] = params[p:]
        else:
            k = np.identity(p)
            k[self.triu] = params
            k[self.tril] = params
        z = self.mx_z
        return z @ k @ z.T

    def calc_k_grad(self, model):
        return self.grad

    def get_bounds(self):
        """
        Return bounding intervals for each of the effects parameters.

        Returns
        -------
        List[tuple]
            List of parameters intervals.

        """
        n = len(self.parameters)
        if self.correlation:
            return [(-1, 1)] * n
        p = self.num_p
        return [(0, None)] * p + [(None, None)] * (n - p)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Statistic and fit indices."""
from .utils import duplication_matrix
from scipy.stats import norm, chi2
from collections import namedtuple
from .optimizer import Optimizer
from .model_generalized_effects import ModelGeneralizedEffects
from .model_effects import ModelEffects
from .model_means import ModelMeans
from .model import Model
import pandas as pd
import numpy as np

ParameterStatistics = namedtuple('ParametersStatistics',
                                 ['value', 'se', 'zscore', 'pvalue'])
SEMStatistics = namedtuple('SEMStatistics', ['dof', 'ml', 'fun',
                                             'chi2', 'dof_baseline',
                                             'chi2_baseline', 'rmsea', 'cfi',
                                             'gfi', 'agfi', 'nfi', 'tli',
                                             'aic', 'bic', 'params'])


def get_baseline_model(model, data=None):
    """
    Retrieve a the baseline model from given model.

    Baseline model here is an independence model where all variables are
    considered to be independent with zero covariance. Only variances are
    estimated.
    Parameters
    ----------
    model : Model
        Model.
    data : pd.DataFrame, optional
        Data, extracting from model will be attempted if no data provided.
        (It's assumed that model.load took place). The default is None.
    cov : pd.DataFrame, optional
        Covariance matrix can be provided in the lack of presense of data.
        The default is None.

    Returns
    -------
    mod : Model
        Baseline model.

    """
    if type(model) is str:
        mod = Model(model, baseline=True)
        if data:
            mod.load(data)
        return mod
    desc = model.description
    tp = type(model)
    tp_name = tp.__name__
    if tp_name not in ('Model', ):
        mod = tp(desc, baseline=True, intercepts=model.intercepts)
    else:
        mod = tp(desc, baseline=True)
    try:
        if not data:
            if hasattr(model, 'mx_data'):
                mx = model.mx_data
                vs = model.vars['observed']
                if hasattr(model, 'mx_g'):
                    mx = np.append(mx, model.mx_g.T, axis=1)
                    vs = vs + model.vars['observed_exogenous']
                data = pd.DataFrame(data=mx, columns=vs)
                if hasattr(model, 'group'):
                    data[model.group] = model.group_data[0, :]
                if tp_name in ('ModelEffects',):
                    mod.load(data, group=model.group, k=model.passed_k)
                else:
                    mod.load(data)
            else:
                data = pd.DataFrame(data=model.mx_cov,
                                    index=model.vars['observed'],
                                    columns=model.vars['observed']) 
                mod.load(cov=data, n_samples=model.n_samples)
    except AttributeError:
        pass
    return mod


def __get_chi2_base(model):
    """
    Calculate chi2 of baseline model.

    Parameters
    ----------
    model : Model
        Model.

    Returns
    -------
    tuple
        chi2, dof of baseline model.

    """
    mod_base = get_baseline_model(model)
    obj = model.last_result.name_obj
    if type(mod_base).__name__ in ('ModelEffects', 'ModelGeneralizedEffects'):
        if obj == 'REML2':
            obj = 'REML'
        elif obj == 'FIML':
            obj = 'ML'
    elif type(mod_base).__name__ in ('ModelMeans', ):
        if obj == 'FIML':
            obj = 'ML'
    mod_base.fit(obj=obj)
    chi2_base = calc_chi2(mod_base)[0]
    return chi2_base, calc_dof(mod_base)


def calc_gfi(model, chi2=None, chi2_base=None):
    """
    Calculate GFI (goodness-of-fit index).

    Parameters
    ----------
    model : model
        Model.
    chi2 : TYPE, optional
        chi2 statistic. The default is None.
    chi2_base : TYPE, optional
        chi2 statistic for baseline model. The default is None.

    Returns
    -------
    float
        GFI.

    """
    if chi2 is None:
        chi2 = calc_chi2(model)[0]
    if chi2_base is None:
        chi2_base = __get_chi2_base(model)[0]
    return 1 - chi2 / chi2_base


def calc_agfi(model, dof=None, dof_bs=None, gfi=None):
    """
    Calculate AGFI (adjusted goodness-of-fit index).

    Parameters
    ----------
    model : Model
        Model.
    dof : int, optional
        Degrees of freedom. The default is None.
    dof_bs : float, optional
        Degrees of freedom of baseline model. The default is None.
    gfi : float, optional
        GFI statistic. The default is None.

    Returns
    -------
    float
        AGFI.

    """
    if dof is None:
        dof = calc_dof(model)
    if dof == 0:
        return np.nan
    if dof_bs is None:
        base = get_baseline_model(model)
        dof_bs = calc_dof(base)
    if gfi is None:
        gfi = calc_gfi(model)
    return 1 - dof_bs / dof * (1 - gfi)  # k * (k + 1) / (2 * dof) * (1 - gfi)


def calc_dof(model):
    """
    Calculate degrees of freedom.

    Parameters
    ----------
    model : Model
        Model.

    Returns
    -------
    int
        DoF.

    """
    p = len(model.vars['observed'])
    return p * (p + 1) // 2 - len(model.param_vals)


def calc_nfi(model, chi2=None, chi2_base=None):
    """
    Calculate NFI (Normed Fit Index).

    Parameters
    ----------
    model : Model
        Model.
    chi2 : float, optional
        chi2 statistic. The default is None.
    chi2_base : TYPE, optional
        chi2 statistic of baseline model. The default is None.

    Returns
    -------
    float
        NFI.

    """
    if chi2 is None:
        chi2 = calc_chi2(model)[0]
    if chi2_base is None:
        chi2_base = __get_chi2_base(model)[0]
    if chi2_base == 0:
        return np.nan
    return (chi2_base - chi2) / chi2_base


def calc_tli(model, dof=None, chi2=None, dof_base=None, chi2_base=None):
    """
    Calculate TLI (Tucker and Lewis Index).

    Parameters
    ----------
    model : Model
        Model.
    dof : int, optional
        Degrees of freedom statistic. The default is None.
    chi2 : float, optional
        chi2 statistic. The default is None.
    dof_base : int, optional
        Degrees of freedom of baseline model. The default is None.
    chi2_base : float, optional
        chi2 statistic of baseline model. The default is None.

    Returns
    -------
    float
        TLI.

    """
    if chi2 is None:
        chi2 = calc_chi2(model)[0]
    if chi2_base is None or dof_base is None:
        chi2_base, dof_base = __get_chi2_base(model)
    if dof is None:
        dof = calc_dof(model)
    if dof == 0 or dof_base == 0:
        return np.nan
    a, b = chi2 / dof, chi2_base / dof_base
    return (b - a) / (b - 1)


def calc_cfi(model, dof=None, chi2=None, dof_base=None, chi2_base=None):
    """
    Calculate CFI (Comparative Fit Index).

    Parameters
    ----------
    model : Model
        Model.
    dof : int, optional
        Degrees of freedom statistic. The default is None.
    chi2 : float, optional
        chi2 statistic. The default is None.
    dof_base : int, optional
        Degrees of freedom statistic of baseline model. The default is None.
    chi2_base : float, optional
        chi2 statistic of baseline model. The default is None.

    Returns
    -------
    float
        CFI.

    """
    if chi2 is None:
        chi2 = calc_chi2(model)[0]
    if dof is None:
        dof = calc_dof(model)
    if chi2_base is None or dof_base is None:
        chi2_base, dof_base = __get_chi2_base(model)
    a = chi2 - dof
    b = chi2_base - dof_base
    return 1 - a / b


def calc_chi2(model, dof=None):
    """
    Calculate chi-square statistic.

    Parameters
    ----------
    model : Model
        Model.
    dof : int, optional
        Degrees of freedom statistic. The default is None.

    Returns
    -------
    tuple
        chi2 statistic and p-value.

    """
    if dof is None:
        dof = calc_dof(model)
    if model.last_result.name_obj == 'FIML':
        stat = model.last_result.fun / model.n_samples
    else:
        stat = model.n_samples * model.last_result.fun
    return stat, 1 - chi2.cdf(stat, dof)


def calc_chi2_sb(model, stat=None, dof=None):
    """
    Calculate scaled (Satorra-Bentler) chi2.

    Parameters
    ----------
    model : Model
        Model.
    stat : float, optional
        chi2 statistic. The default is None.
    dof : int, optional
        Degrees of freedom. The default is None.

    Returns
    -------
    float
        SB chi2.
    float
        p-value.

    """
    if stat is None:
        stat = calc_chi2(model)[0]
    if dof is None:
        dof = calc_dof(model)
    sigma, (m, c) = model.calc_sigma()
    inds = np.triu_indices_from(sigma)
    grad = np.array([g[inds].flatten(order='F')
                    for g in model.calc_sigma_grad(m, c)]).T
    sigma_inv = np.linalg.inv(sigma)
    k = np.kron(sigma_inv, sigma_inv)
    dup = duplication_matrix(sigma.shape[0])
    w = dup.T @ k @ dup / 2
    w = np.linalg.inv(w)
    t = w @ grad
    u = w - t @ np.linalg.inv(grad.T @ t) @ t.T
    if hasattr(model, 'mx_w'):
        c = np.einsum('ij,ji->', u, model.mx_w) / dof
    else:
        c = np.trace(u @ np.linalg.inv(w)) / dof
        return t, grad, w, u
    stat /= c
    return stat, 1 - chi2.cdf(stat, dof)


def calc_rmsea(model, chi2=None, dof=None):
    """
    Calculate RMSEA statistic.

    Parameters
    ----------
    model : Model
        Model.
    chi2 : float, optional
        chi2 statistic. The default is None.
    dof : int, optional
        Degrees of freedom statistic. The default is None.

    Returns
    -------
    float
        RMSEA.

    """
    if chi2 is None:
        chi2 = calc_chi2(model)[0]
    if dof is None:
        dof = calc_dof(model)
    if chi2 < dof:
        return 0
    return np.sqrt((chi2 / dof - 1) / (model.n_samples - 1))


def calc_likelihood(model):
    """
    Calculate likelihood.

    Parameters
    ----------
    model : Model
        Model.

    Returns
    -------
    float
        Loglikelihood.

    """
    # TODO
    return model.obj_mlw(model.param_vals)


def calc_aic(model, lh=None):
    """
    Calculate Akaike's Information Criteria (AIC) statistic.

    Parameters
    ----------
    model : Model
        Model.
    lh : float, optional
        Loglikelihood. The default is None.

    Returns
    -------
    float
        AIC.

    """
    if lh is None:
        lh = calc_likelihood(model)
    return 2 * (len(model.param_vals) - lh)


def calc_bic(model, lh=None):
    """
    Calculate Bayesian Information Criteria (BIC) statistic.

    Parameters
    ----------
    model : Model
        Model.
    lh : float, optional
        Loglikelihood. The default is None.

    Returns
    -------
    float
        BIC.

    """
    if lh is None:
        lh = calc_likelihood(model)
    k, n = len(model.param_vals), model.n_samples
    return np.log(n) * k - 2 * lh


def calc_se(model, information='expected', robust=False):
    """
    Calculate standard errors.

    Parameters
    ----------
    model : Model
        Model.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. The default is 'expected'.
    robust : bool, optional
        If True, then robust SE are computed instead. Robustness here means
        that MLR-esque sandwich correction is applied. The default is False.

    Returns
    -------
    list
        list of standard errors of model's active parameters.

    """
    if robust or information != 'expected':
        # An explanation of "mult" for a reader:
        # When I was writing code and deriving some formulas for different
        # objective functions, I was finding myself quiet often dropping
        # constants when they appeared to be not necessary. Sometimes I didn't.
        # It is a mess. Here I am taking that in account. Some may say that
        # that I'd better just get those goddamn constants back in place
        # and get rid of this clusterfuck, and they'd be right, but I am
        # currently too hesitant to do it myself. 
        # My sincere apologies.
        # Georgy
        mult = model.n_samples / 2
        if hasattr(model, 'last_result'):
            fun, grad = model.get_objective(model.last_result.name_obj)
            if model.last_result.name_obj in ('FIML'):
                mult = 0.5
        else:
            try:
                fun, grad = model.get_objective('MatNorm')
            except KeyError:
                try:
                    fun, grad = model.get_objective('FIML')
                except KeyError:
                    fun, grad = model.get_objective('MLW')
                    mult = 1.0
        
    if information == 'expected':
        mx_inf = model.calc_fim(inverse=True)[1]
    else:
        from numdifftools import Gradient, Hessian
        if grad is None:
            hess = Hessian(fun)(model.param_vals)
        else:
            hess = Gradient(grad)(model.param_vals)
        mx_inf = np.linalg.pinv(hess) / mult
    if robust:
        g = sum(np.outer(g,g) for g in model.grad_se_g(model.param_vals))
        mx_inf = mx_inf @ g @ mx_inf
    variances = mx_inf.diagonal().copy()
    inds = (variances < 0) & (variances > -1e-1)
    variances[inds] = 1e-12
    variances[variances < 0] = np.nan  # So numpy won't throw a warning.
    return np.sqrt(variances) 


def calc_zvals(model, std_errors=None, information='expected'):
    """
    Calculate Z-score.

    Parameters
    ----------
    model : Model
        Model.
    std_errors : list, optional
        Standard errors. The default is None.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. No use if std_errors are provided.
        The default is 'expected'.

    Returns
    -------
    list
        Z-scores.

    """
    if std_errors is None:
        std_errors = calc_se(model, information=information)
    return [val / (std + 1e-12) for val, std in zip(list(model.param_vals),
                                                    std_errors)]


def calc_pvals(model, z_scores=None, information='expected'):
    """
    Calculate p-values.

    Parameters
    ----------
    model : Model
        Model.
    z_scores : list, optional
        Z-sores. The default is None.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. No use if std_errors are provided.
        The default is 'expected'.

    Returns
    -------
    list
        P-values.

    """
    if z_scores is None:
        z_scores = calc_zvals(model, information=information)
    return [2 * (1 - norm.cdf(abs(z))) for z in z_scores]


def calc_stats(model):
    """
    Calculate fit indices and other auxiliary statistics.

    Parameters
    ----------
    model : Model
        Fit model instance.

    Returns
    -------
    pd.DataFrame
        Fit indices and statistics.

    """
    if not hasattr(model, 'last_result'):
        raise Exception('Can''t gather statitics from model until it is fit.')
    try:
        lh = calc_likelihood(model)
    except np.linalg.LinAlgError:
        lh = np.nan
    aic = calc_aic(model, lh)
    bic = calc_bic(model, lh)
    dof = calc_dof(model)
    chi2_base, dof_base = __get_chi2_base(model)
    chi2 = calc_chi2(model, dof)
    rmsea = calc_rmsea(model, chi2[0], dof)
    cfi = calc_cfi(model, dof, chi2[0], dof_base, chi2_base)
    gfi = calc_gfi(model, chi2[0], chi2_base)
    agfi = calc_agfi(model, dof, dof_base, gfi)
    nfi = calc_nfi(model, chi2[0], chi2_base)
    tli = calc_tli(model, dof, chi2[0], dof_base, chi2_base)
    d = {'DoF': dof, 'DoF Baseline': dof_base,
         'chi2': chi2[0], 'chi2 p-value': chi2[1], 'chi2 Baseline': chi2_base,
         'CFI': cfi, 'GFI': gfi, 'AGFI': agfi, 'NFI': nfi, 'TLI': tli,
         'RMSEA': rmsea, 'AIC': aic, 'BIC': bic, 'LogLik': lh}
    return pd.DataFrame(d, index=['Value'])


def gather_statistics(model, information='expected', robust=False):
    """
    Retrieve all statistics as specified in SEMStatistics structure.

    Needed for backwards-compaitability with semopy 1.+.
    Parameters
    ----------
    model : Model
        Model.
    information : str
        If 'expected', expected Fisher information is used. Otherwise,
        observed information is employed. The default is 'expected'.
    robust : bool, optional
        If True, then robust SE are computed instead. Robustness here means
        that MLR-esque sandwich correction is applied. The default is False.

    Raises
    ------
    Exception
        Rises when no data is loaded into the model.

    Returns
    -------
    SEMStatistics
        Namedtuple containing all statistics available to semopy.

    """
    if isinstance(model, Optimizer):
        model = model.model
    if not hasattr(model, 'last_result'):
        raise Exception('Can''t gather statitics from model until it is fit.')
    values = model.param_vals.copy()
    try:
        std_errors = calc_se(model, information=information, robust=robust)
    except np.linalg.LinAlgError:
        std_errors = np.array([np.nan] * len(values))
    z_scores = calc_zvals(model, std_errors)
    pvalues = calc_pvals(model, z_scores)
    try:
        lh = calc_likelihood(model)
    except np.linalg.LinAlgError:
        lh = np.nan
    aic = calc_aic(model, lh)
    bic = calc_bic(model, lh)
    paramStats = [ParameterStatistics(val, std, ztest, pvalue)
                  for val, std, ztest, pvalue
                  in zip(values, std_errors, z_scores, pvalues)]
    dof = calc_dof(model)
    chi2_base, dof_base = __get_chi2_base(model)
    chi2 = calc_chi2(model, dof)
    rmsea = calc_rmsea(model, chi2[0], dof)
    cfi = calc_cfi(model, dof, chi2[0], dof_base, chi2_base)
    gfi = calc_gfi(model, chi2[0], chi2_base)
    agfi = calc_agfi(model, dof, dof_base, gfi)
    nfi = calc_nfi(model, chi2[0], chi2_base)
    tli = calc_tli(model, dof, chi2[0], dof_base, chi2_base)
    return SEMStatistics(dof, model.last_result.fun, lh, chi2, dof_base,
                         chi2_base, rmsea, cfi, gfi, agfi, nfi, tli, aic, bic,
                         paramStats)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains functions for stating values estimation."""
from scipy.stats import linregress
import numpy as np


def start_beta(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Parameters in beta are traditionally set to 0 at start.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    return 0.0


def start_lambda(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Manifest variables are regressed onto their counterpart with fixed
    regression coefficient.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    if rval not in model.vars['latent']:
        return 0.0
    obs = model.vars['observed']
    first = rval
    while first not in obs:
        try:
            first = model.first_manifs[first]
            if first == rval:
                return 0.0
        except KeyError:
            return 0.0
    if first is None or not hasattr(model, 'mx_data'):
        return 0.0
    i, j = obs.index(first), obs.index(lval)
    data = model.mx_data
    x, y = data[:, i], data[:, j]
    mask = np.isfinite(x) & np.isfinite(y)
    return linregress(x[mask], y[mask]).slope


def start_psi(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Exogenous covariances are fixed to their empirical values.
    All other variances are halved. Latent variances are set to 0.05,
    everything else is set to zero.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    lat = model.vars['latent']
    if rval in lat or lval in lat:
        if rval == lval:
            return 0.05
        return 0.0
    exo = model.vars['exogenous']
    obs = model.vars['observed']
    i, j = obs.index(lval), obs.index(rval)
    if lval in exo:
        return model.mx_cov[i, j]
    elif i == j:
        return model.mx_cov[i, j] / 2
    return 0.0


def start_theta(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Variances are set to half of observed variances.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    if lval != rval:
        return 0.0
    obs = model.vars['observed']
    i, j = obs.index(lval), obs.index(rval)
    return model.mx_cov[i, j] / 2


def start_gamma1(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Parameters in Gamma1 are set to 0 at start unless we are dealing with
    means, then they are estimated as means.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    if rval == '1':
        mx = model.mx_data
        i = model.vars['observed'].index(lval)
        return np.nanmean(mx[:, i]) / 2
    return 0.0


def start_gamma2(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Parameters in Gamma2 are set to 0 at start unless we are dealing with
    means, then they are estimated as means.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    if rval == '1':
        mx = model.mx_data
        i = model.vars['observed'].index(lval)
        return np.nanmean(mx[:, i]) /  2
    return 0.0

def start_d(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    In future a sophisticated procedure will be provided.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    if lval == rval:
        v = model.effects_loadings.get(lval, 0.1) / 2
        return v
    return 0.0

def start_v(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    return 1.0
  

'''---------------------------------IMPUTER---------------------------------'''


def start_data_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just calculates mean.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    obs = model.mod.vars['observed']
    try:
        i = obs.index(rval)
    except ValueError:
        return 0.0
    mx = model.mod.mx_data
    return np.nanmean(mx[:, i])

def start_g_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just calculates mean.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    obs = model.mod.vars['observed_exogenous']
    try:
        i = obs.index(rval)
    except ValueError:
        return 0.0
    mx = model.mod.mx_g1
    return np.nanmean(mx[i, :])


def start_beta_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original model.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_beta
    rows, cols = model.mod.names_beta
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


def start_gamma1_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original model.
    Parameters
    ----------
    model : Model
        Imputer instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_gamma1
    rows, cols = model.mod.names_gamma1
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


def start_gamma2_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original model.
    Parameters
    ----------
    model : Model
        Imputer instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_gamma2
    rows, cols = model.mod.names_gamma2
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


def start_lambda_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original data.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_lambda
    rows, cols = model.mod.names_lambda
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


def start_psi_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original data.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_psi
    rows, cols = model.mod.names_psi
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


def start_theta_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original data.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_theta
    rows, cols = model.mod.names_theta
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v

def start_d_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original data.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_d
    rows, cols = model.mod.names_d
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v

def start_v_imp(model, lval: str, rval: str):
    """
    Calculate starting value for parameter in data given data in model.

    For Imputer -- just copies values from original data.
    Parameters
    ----------
    model : Model
        Model instance.
    lval : str
        L-value name.
    rval : str
        R-value name.

    Returns
    -------
    float
        Starting value.

    """
    mx = model.mod.mx_v
    rows, cols = model.mod.names_v
    i, j = rows.index(lval), cols.index(rval)
    v = mx[i, j]
    return v


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Parser module."""
from collections import defaultdict, namedtuple
import re

Operation = namedtuple('Operation', 'name, params, onto',
                       defaults=(None, None))
__prt_lvalue = r'(\w[\w\.]*(?:\s*,\s*[\w.]*)*)'
__prt_op = r'\s*((?:\s\w+\s)|(?:[=~\\\*@\$<>\-]+\S*?))\s*'
__prt_rvalue = r'(-?\w[\w.]*(?:\s*\*\s*\w[\w.]*)?(?:\s*\+\s*-?\w[\w.]*(?:\s*\*\s*\w[\w.]*)?)*)'
PTRN_EFFECT = re.compile(__prt_lvalue + __prt_op + __prt_rvalue)
PTRN_OPERATION = re.compile(r'([A-Z][A-Z_]+(?:\(.*\))?)\s*([\w\s]+)*')
PTRN_RVALUE = re.compile(r'((-?\w[\w.]*\*)?\w[\w.]*)')
PTRN_OP = re.compile(r'(\w+)(\(.*\))?')


def separate_token(token: str):
    """
    Test if token satisfies basic command semopy syntax and separates token.

    Parameters
    ----------
    token : str
        A text line with either effect command or operation command.

    Raises
    ------
    SyntaxError
        Token happens to be incorrect, i.e. it does not follows basic
        semopy command pattern.

    Returns
    -------
    bool
        True if token is an effect command.
    tuple
        A tuple of (lvalue, operation, rvalue) if command is effect or
        (operation, operands) if command is operation.

    """
    effect = PTRN_EFFECT.fullmatch(token)
    if effect:
        return True, effect.groups()
    operation = PTRN_OPERATION.fullmatch(token)
    if not operation:
        raise SyntaxError(f'Invalind syntax for line:\n{token}')
    return False, operation.groups()


def parse_rvalues(token: str):
    """
    Separate token by  '+' sign and parses expression "val*x" into tuples.

    Parameters
    ----------
    token : str
        Right values from operand.

    Raises
    ------
    Exception
        Rises when a certain rvalue can't be processed.

    Returns
    -------
    rvalues : dict
        A mapping Variable->Multiplicator.

    """
    token = token.replace(' ', '')
    rvalues = dict()
    for tok in token.split('+'):
        rval = PTRN_RVALUE.match(tok)
        if not rval:
            raise Exception(f'{rval} does not seem like a correct semopy \
                            expression')
        groups = rval.groups()
        name = groups[0].split('*')[-1]
        rvalues[name] = groups[1][:-1] if groups[1] else None
    return rvalues


def parse_operation(operation: str, operands: str):
    """
    Parse an operation according to semopy syntax.

    Parameters
    ----------
    operation : str
        Operation string with possible arguments.
    operands : str
        Variables/values that operation acts upon.

    Raises
    ------
    SyntaxError
        Rises when there is an error during parsing.

    Returns
    -------
    operation : Operation
        Named tuple containing information on operation.

    """
    oper = PTRN_OP.match(operation)
    if not oper:
        raise SyntaxError(f'Incorrect operation pattern: {operation}')
    operands = [op.strip() for op in operands.split()] if operands else list()
    groups = oper.groups()
    name = groups[0]
    params = groups[1]
    if params is not None:
        params = [t.strip() for t in params[1:-1].split(',')]
    operation = Operation(name, params, operands)
    return operation


def parse_desc(desc: str):
    """
    Parse a model description provided in semopy's format.

    Parameters
    ----------
    desc : str
        Model description in semopy format.

    Returns
    -------
    effects : defaultdict
        Mapping operation->lvalue->rvalue->multiplicator.
    operations : dict
        Mapping operationName->list[Operation type].

    """
    desc = desc.replace(chr(8764), chr(126))
    effects = defaultdict(lambda: defaultdict(dict))
    operations = defaultdict(list)
    for line in desc.splitlines():
        try:
            i = line.index('#')
            line = line[:i]
        except ValueError:
            pass
        line = line.strip()
        if line:
            is_effect, items = separate_token(line)
            if is_effect:
                lefts, op_symb, rights = items
                for left in lefts.split(','):
                    rvalues = parse_rvalues(rights)
                    effects[op_symb][left.strip()].update(rvalues)
            else:
                operation, operands = items
                t = parse_operation(operation, operands)
                operations[t.name].append(t)
    return effects, operations



'''Create fancy HTML report for a given fitted model.'''
import logging
from .. import semplot, calc_stats, __version__
from collections import defaultdict
import pandas as pd
import shutil
import os

__folder = os.path.dirname(os.path.abspath(__file__))
__index = os.path.join(__folder, 'report.html')
__vis = os.path.join(__folder, 'vis.txt')
__css = os.path.join(os.path.join(__folder, 'css'), 'bootstrap.min.css')
__js = os.path.join(os.path.join(__folder, 'js'), 'bootstrap.min.js')
__navitem = '''<li class="nav-item" role="presentation">
<button class="nav-link {2}" id="{0}-tab" data-bs-toggle="tab" data-bs-target="#{0}" type="button" role="tab" aria-controls="{0}" aria-selected="{3}">{1}</button>
</li>'''
__tab = '''<div class="tab-pane fade {2}" id="{0}" role="tabpanel" aria-labelledby="{0}-tab">{1}</div>'''
__start = '''<ul class="nav nav-tabs" id="{0}" role="tablist">'''
__cont = '''</ul><div class="tab-content" id="{0}Content">'''
__end = '''</div>'''

def report(model, name: str, path='', std_est=False, se_robust=False,
           **kwargs):
    """
    Generate a report in the form of HTML file.

    Parameters
    ----------
    model : ModelBase
        Arbitrary fitted model instance.
    name : str
        A name that user gives to the model. Can be thought of as a title. Will
        be used as a directory name.
    path : str, optional
        Path to folder where report will be stored. The default is ''.
    std_est : bool, optional
        If True, then standardized estimates are used. The default is False.
    se_robust : TYPE, optional
        If True, then robust p-values are calculated. The default is False.
    **kwargs : dict
        Extra arguments for semplot.

    Returns
    -------
    None.

    """
    desc = model.description
    n_samples = model.n_samples
    model_name = type(model).__name__
    optresult = str(model.last_result)
    if model.last_result.success:
        color = 'success'
        converged = 'Yes'
    else:
        color = 'danger'
        converged = 'No'
    obj = model.last_result.name_obj
    clasv = str()
    pt = '<div class="row"><div class="col-sm-2">{}</div>'\
         '<div class="col-sm-6">{}</div></div>'
    for c, its in model.vars.items():
        clasv += pt.format(c, ', '.join(its))
    ins = model.inspect(std_est=std_est, se_robust=se_robust)
    measurement = defaultdict(list)
    reg = defaultdict(list)
    intercepts = defaultdict(list)
    var = defaultdict(list)
    cov = defaultdict(list)
    other = defaultdict(lambda: defaultdict(list))
    ests = str()
    outs = model.vars['_output']
    obs = model.vars['observed']
    lats = model.vars['latent']
    out_obs = set(outs) & set(obs)
    for _, row in ins.iterrows():
        lval = row['lval']
        op = row['op']
        rval = row['rval']
        pval = row['p-value']
        est = row['Estimate']
        try:
            pval = '{:.3f}'.format(float(pval))
        except ValueError:
            pass
        try:
            est = '{:.3f}'.format(float(est))
        except ValueError:
            pass
        if std_est:
            est_std = row['Est. Std']
            try:
                est_std = '{:.3f}'.format(float(est_std))
            except ValueError:
                pass
            t = [rval, est, est_std, pval]
        else:
            t = [rval, est, pval]
        if op == '~':
            if rval in (1, '1'):
                intercepts[lval].append(t)
            elif (rval in lats) and (lval in out_obs):
                t[0] = lval
                measurement[rval].append(t)
            else:
                reg[lval].append(t)
        elif op == '~~':
            if lval != rval:
                cov[lval].append(t)
            else:
                var[lval].append(t)
        else:
            other[op][lval].append(t)
    pt = "<div class='col-sm-2'>{}</div><div class='col-sm-2'>{}</div>"\
         "<div class='col-sm-1'>{}</div><div class='col-sm-1'>{}</div>"
    if std_est:
        pt += "<div class='col-sm-1'>{}</div>"
        t = pt.format('', '', 'Estimate', 'Std. Estimate',
                      'P-value')
    else:
        t = pt.format('', '', 'Estimate', 'P-value')
    ests = '<div class="row">' + t + '</div>'
    ptt = "<div class='col-sm-2 text-end'>{}</div>"\
          "<div class='col-sm-2'>{}</div><div class='col-sm-1'>{}</div>"\
          "<div class='col-sm-1'>{}</div>"
    if std_est:
        ptt += "<div class='col-sm-1'>{}</div>"
    if measurement:
        if std_est:
            t = pt.format('Measurement:', '', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in measurement.items():
                t = ptt.format(lval + ' =~', '', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, est_std, pval) in its:
                    t = ptt.format('', rval, est, est_std, pval)
                    ests += '<div class="row">' + t + '</div>'
        else:
            ests += pt.format('Measurement:', '', '', '')
            for lval, its in measurement.items():
                t = ptt.format(lval + ' =~', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, pval) in its:
                    t = ptt.format('', rval, est, pval)
                    ests += '<div class="row">' + t + '</div>'
    if reg:
        if std_est:
            t = pt.format('Regressions:', '', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in measurement.items():
                t = ptt.format(lval + ' ~', '', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, est_std, pval) in its:
                    t = ptt.format('', rval, est, est_std, pval)
                    ests += '<div class="row">' + t + '</div>'
        else:
            ests += pt.format('Regressions:', '', '', '')
            for lval, its in reg.items():
                t = ptt.format(lval + ' ~', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, pval) in its:
                    t = ptt.format('', rval, est, pval)
                    ests += '<div class="row">' + t + '</div>'
    if intercepts:
        if std_est:
            t = pt.format('Intercepts:', '', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in intercepts.items():
                for (rval, est, est_std, pval) in its:
                    t = ptt.format('', lval, est, est_std, pval)
                    ests += '<div class="row">' + t + '</div>'
        else:
            ests += pt.format('Intercepts:', '', '', '')
            for lval, its in intercepts.items():
                for (rval, est, pval) in its:
                    t = ptt.format('', lval, est, pval)
                    ests += '<div class="row">' + t + '</div>'
                    
    if var:
        if std_est:
            t = pt.format('Variances:', '', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in var.items():
                for (rval, est, est_std, pval) in its:
                    t = ptt.format('', rval, est, est_std, pval)
                    ests += '<div class="row">' + t + '</div>'
        else:
            ests += pt.format('Variances:', '', '', '')
            for lval, its in var.items():
                for (rval, est, pval) in its:
                    t = ptt.format('', rval, est, pval)
                    ests += '<div class="row">' + t + '</div>'
    if cov:
        if std_est:
            t = pt.format('Covariances:', '', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in cov.items():
                t = ptt.format(lval + ' ~', '', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, est_std, pval) in its:
                    t = ptt.format('', rval, est, est_std, pval)
                    ests += '<div class="row">' + t + '</div>'
        else:
            t = pt.format('Covariances:', '', '', '')
            ests += '<div class="row">' + t + '</div>'
            for lval, its in cov.items():
                t = ptt.format(lval + ' ~', '', '', '')
                ests += '<div class="row">' + t + '</div>'
                for (rval, est, pval) in its:
                    t = ptt.format('', rval, est, pval)
                    ests += '<div class="row">' + t + '</div>'
    if other:
        if std_est:
            t = pt.format('Other operators:', '', '', '', '')
        else:
            t = pt.format('Other operators:', '', '', '')
        ests += '<div class="row">' + t + '</div>'
        for op, rvals in other.items():
            if std_est:
                for lval, its in rvals.items():
                    t = ptt.format(lval + f' {op}', '', '', '', '')
                    ests += '<div class="row">' + t + '</div>'
                    for (rval, est, est_std, pval) in its:
                        t = ptt.format('', rval, est, est_std, pval)
                        ests += '<div class="row">' + t + '</div>'
            else:
                for lval, its in rvals.items():
                    t = ptt.format(lval + f' {op}', '', '', '')
                    ests += '<div class="row">' + t + '</div>'
                    for (rval, est, pval) in its:
                        t = ptt.format('', rval, est, pval)
                        ests += '<div class="row">' + t + '</div>'
    path = os.path.join(path, name)
    if not os.path.isdir(path):
        os.mkdir(path)
    try:
        pp = os.path.join(path, 'plots')
        if not os.path.isdir(pp):
            os.mkdir(pp)
        semplot(model, os.path.join(pp, '1.png'), plot_ests=False, **kwargs)
        semplot(model, os.path.join(pp, '2.png'), plot_ests=True, **kwargs)
        semplot(model, os.path.join(pp, '3.png'), plot_ests=False,
                plot_covs=True, **kwargs)
        semplot(model, os.path.join(pp, '4.png'), plot_ests=True,
                       plot_covs=True, **kwargs)
        with open(__vis, 'r') as f:
            vis = f.read()
    except Exception as e:
        s = 'Could not plot model. Possible Graphviz installation issues.' 
        logging.warning(s + ' ' + str(e))
        vis = f'<div class="alert alert-warning" role="alert">{s}</div>'
    stats = calc_stats(model)
    fitindices = str()
    for i in range(0, stats.shape[1], 5):
        fitindices += '<table class="table">'
        a = "<thead><tr>"
        b = "<tbody><tr>"
        for j in range(i, min(stats.shape[1], i + 5)):
            a += "<th scope='col'>{}</th>".format(stats.columns[j])
            t = stats[stats.columns[j]].values.flatten()[0]
            try:
                t = '{:.3f}'.format(t)
            except ValueError:
                pass
            b += "<td>{}</td>".format(t)
        b += "</tr></tbody>"
        a += "</tr></thead>"
        fitindices += a + b + '</table>'
    pt = '<center><h4>{}</h4></center>{}'
    fmt = lambda x: '{:.2f}'.format(x)
    c = 'table table-bordered'
    d = defaultdict(list)
    for n, mx in model.inspect('mx').items():
        d[n].append(mx.to_html(float_format=fmt, classes=c, 
                               justify='center'))
    for n, mx in model.inspect('mx', what='start').items():
        d[n].append(mx.to_html(float_format=fmt, classes=c, 
                                justify='center'))
    for n, mx in model.inspect('mx', what='name').items():
        d[n].append(mx.to_html(float_format=fmt, classes=c, 
                               justify='center'))
    obs = model.vars['observed']
    sigma, _ = model.calc_sigma()
    cov = pd.DataFrame(model.mx_cov, columns=obs, index=obs)
    cov = cov.to_html(float_format=fmt, classes=c,justify='center')
    if hasattr(model, 'calc_l'):
        sigma = model.calc_l(sigma=sigma)
    sigma = pd.DataFrame(sigma, columns=obs, index=obs)
    sigma = sigma.to_html(float_format=fmt, classes=c,justify='center')
    matrices = '<center><h4>Covariance matrix</h4></center>\n'
    matrices += __start.format('cov')
    matrices += __navitem.format('sample', 'Sample', 'active', 'true')
    matrices += __navitem.format('modelimplied', 'Model-implied', '', 'false')
    matrices += __cont.format('cov')
    matrices += __tab.format('sample', cov, 'show active')
    matrices += __tab.format('modelimplied', sigma, '')
    matrices += __end
    matrices += '<br><center><h4>Model matrices</h4></center>'
    matrices += __start.format('modelmatrices')
    for i, n in enumerate(d.keys()):
        if i == 0:
            matrices += __navitem.format(f'mx{n}', n, 'active', 'true')
        else:
            matrices += __navitem.format(f'mx{n}', n, '', 'false')
    matrices += __cont.format('cov')
    for i, (n, mxs) in enumerate(d.items()):
        sub = __start.format(n + 'sub')
        for j, (mode, mx) in enumerate(zip(('Estimates', 'Start', 'Names'),
                                         mxs)):
            if j == 0:
                sub += __navitem.format(n + mode, mode, 'active', 'true')
            else:
                sub += __navitem.format(n + mode, mode, '', 'false')
        sub += __cont.format(name + 'sub')
        for j, (mode, mx) in enumerate(zip(('Estimates', 'Start', 'Names'),
                                         mxs)):
            if j == 0:
                sub += __tab.format(n + mode, mx, 'show active')
            else:
                sub += __tab.format(n + mode, mx, '')
        sub += __end
        if i == 0:
            matrices += __tab.format(f'mx{n}', sub,'show active')
        else:
            matrices += __tab.format(f'mx{n}', sub, '')
    matrices += __end
        
    with open(__index, 'r') as f:
        html = f.read()
    html = html.format(ModelName=name, OptimizerReport=optresult,
                       ModelDescription=desc, VariableClassification=clasv,
                       NSamples=n_samples, ConvColor=color,
                       Converged=converged, Objective=obj,
                       Visualization=vis, Estimates=ests, Matrices=matrices,
                       InspectionTable=str(ins), FitIndices=fitindices,
                       Version=__version__,  ModelClass=model_name)
    
    with open(os.path.join(path, 'report.html'), 'w') as f:
        f.write(html)
    css = os.path.join(path, 'css')
    if not os.path.isdir(css):
        os.mkdir(css)
    shutil.copy(__css, css)
    js = os.path.join(path, 'js')
    if not os.path.isdir(js):
        os.mkdir(js)
    shutil.copy(__js, js)



# -*- coding: utf-8 -*-
"""Backwards-compatibility module for semopy 1.+."""
from .model import Model
from copy import deepcopy


class Optimizer():
    def __init__(self, mod: Model):
        if not hasattr(mod, 'mx_data'):
            raise Exception('load_dataset method must be called prior to'
                            ' passing Model to an Optimizer.')
        self.model = deepcopy(mod)

    def optimize(self, obj='MLW', method='SLSQP'):
        res = self.model.fit(obj=obj, solver=method)
        return res.fun




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Mean estimator for Model without meanstructure. FOR INTERNAL USAGE ONLY."""

import pandas as pd
import numpy as np
from .model import Model
from .solver import Solver
from .utils import chol_inv
from .stats import calc_se, calc_zvals, calc_pvals


class MeanEstimator():
    """Simple mean estimator for the Model without meanstructure."""

    __slots__ = ['sigma', 'm', 'mx_data', 'mx_data_t', 'diffs', 'range_mu',
                 'range_vu', 'mu', 'vu', 'param_vals', 'last_result',
                 'sigma_inv', 'vars']

    def __init__(self, model: Model):
        """
        Instantiate mean estimator.

        Parameters
        ----------
        model : Model
            Model without meanstructure.

        Raises
        ------
        Exception
            Raises when no data is available.

        Returns
        -------
        None.

        """
        if not hasattr(model, 'mx_data'):
            raise Exception('Can''t estimate means if no data is available.')
        self.sigma, (self.m, _) = model.calc_sigma()
        self.sigma_inv = chol_inv(self.sigma)
        self.mx_data = model.mx_data
        self.mx_data_t = model.mx_data.T
        self.vars = dict()
        obs = model.vars['observed']
        self.vars['observed'] = obs
        set_obs = set(obs)
        inner = model.vars['inner']
        outer = model.vars['_output']
        self.vars['inner'] = inner
        self.vars['_output'] = outer
        self.mu = np.zeros((len(inner), 1))
        self.vu = np.zeros((len(obs), 1))
        obs_inner = set_obs & set(inner)
        inner = [i for i, v in enumerate(inner) if v in set_obs]
        innert = [obs.index(v) for v in obs if v in obs_inner]
        outer = [i for i, v in enumerate(obs) if v in outer]
        self.mu[inner, 0] = np.nanmean(self.mx_data_t[:, innert], axis=0)
        self.vu[outer, 0] = np.nanmean(self.mx_data_t[:, outer], axis=0)
        self.range_mu = (inner, (0, len(inner)))
        self.range_vu = (outer, (len(inner), len(inner) + len(outer)))
        self.build_diff_vectors()
        self.param_vals = list(self.mu[inner].flatten())
        self.param_vals += list(self.vu[outer].flatten())
        self.param_vals = np.array(self.param_vals)

    def build_diff_vectors(self):
        """
        Calculate derivatives of mean vectors.

        Returns
        -------
        None.

        """
        diffs = list()
        mt = -self.m
        for i in self.range_mu[0]:
            diffs.append(mt[:, i][:, np.newaxis])
        for i in self.range_vu[0]:
            t = np.zeros_like(self.vu)
            t[i, 0] = -1
            diffs.append(t)
        self.diffs = diffs

    def fit(self, obj='ML', solver='SLSQP'):
        """
        Estimate means.

        Parameters
        ----------
        obj : str, optional
            Objective function to minimize. Possible values are 'ML', 'ULS'.
            'GLS' is same as 'ML'. The default is 'MLW'.
        solver : TYPE, optional
            Optimizaiton method. Currently scipy-only methods are available.
            The default is 'SLSQP'.

        Raises
        ------
        Exception
            Rises when attempting to use FIML in absence of full data.

        Returns
        -------
        Optimization result.

        """
        fun, grad = self.get_objective(obj)
        solver = Solver(solver, fun, grad, self.param_vals)
        res = solver.solve()
        res.name_obj = obj
        self.param_vals = res.x
        self.update_vectors(res.x)
        self.last_result = res
        return res

    def get_objective(self, name: str):
        """
        Retrieve objective function and its gradient by name.

        Parameters
        ----------
        name : str
            Name of objective function.

        Raises
        ------
        KeyError
            Rises if incorrect name is provided.

        Returns
        -------
        tuple
            Objective function and gradient function.

        """
        d = {'ML': (self.obj_ml, self.grad_ml),
             'GLS': (self.obj_ml, self.grad_ml),
             'ULS': (self.obj_uls, self.grad_uls)}
        try:
            return d[name]
        except KeyError:
            raise KeyError(f'{name} is unknown objective function.')

    def update_vectors(self, x: np.ndarray):
        """
        Update mean vectors from vector of parameters.

        Parameters
        ----------
        x : np.ndarray
            DESCRIPTION.

        Returns
        -------
        None.

        """
        if self.range_mu[0]:
            self.mu[self.range_mu[0], 0] = x[:self.range_mu[1][1]]
        if self.range_vu[0]:
            self.vu[self.range_vu[0], 0] = x[self.range_vu[1][0]:]

    def calc_center(self):
        """
        Calculate centered data.

        Returns
        -------
        np.ndarray
            Data with means subtracted.

        """
        return self.mx_data_t - self.m @ self.mu - self.vu

    def calc_center_grad(self):
        """
        Calculate center gradient.

        Returns
        -------
        np.ndarray
            Centered gradient.

        """
        return self.diffs

    def calc_fim(self, inverse=False):
        """
        Calculate Fisher Information Matrix.

        Exponential-family distributions are assumed.
        Parameters
        ----------
        inverse : bool, optional
            If True, function also returns inverse of FIM. The default is
            False.

        Returns
        -------
        np.ndarray
            FIM.
        np.ndarray, optional
            FIM^{-1}.

        """
        center_grad = self.calc_center_grad()
        sz = len(center_grad)
        info = np.zeros((sz, sz))
        sgs = [sg.T @ self.sigma_inv for sg in center_grad]
        for i in range(sz):
            for k in range(i, sz):
                info[i, k] = np.einsum('ij,ji->', sgs[i], center_grad[k])
        fim = info + np.triu(info, 1).T
        if inverse:
            fim_inv = chol_inv(fim)
            return (fim, fim_inv)
        return fim

    def obj_ml(self, x: np.ndarray):
        """
        Calculate gaussian maximum likelihood objective function.

        Here it is the same as GLS.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood value.

        """
        self.update_vectors(x)
        center = self.calc_center()
        return np.einsum('ij,ji->', center.T, self.sigma_inv @ center)

    def grad_ml(self, x: np.ndarray):
        """
        Calculate maximum likelihood objective function gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Array of derivatives at point x .

        """
        self.update_vectors(x)
        t = self.calc_center().T @ self.sigma_inv
        res = np.zeros(x.shape)
        df = self.calc_center_grad()
        for i in range(len(res)):
            res[i] = 2 * np.einsum('ij,ji->', t, df[i])
        return res

    def obj_uls(self, x: np.ndarray):
        """
        Calculate gaussian maximum likelihood objective function.

        Here it is the same as GLS.
        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        float
            Loglikelihood value.

        """
        self.update_vectors(x)
        center = self.calc_center()
        return np.einsum('ij,ji->', center.T, center)

    def grad_uls(self, x: np.ndarray):
        """
        Calculate ULS objective function gradient.

        Parameters
        ----------
        x : np.ndarray
            Parameters vector.

        Returns
        -------
        np.ndarray
            Array of derivatives at point x .

        """
        self.update_vectors(x)
        t = self.calc_center().T
        res = np.zeros(x.shape)
        df = self.calc_center_grad()
        for i in range(len(res)):
            res[i] = 2 * np.einsum('ij,ji->', t, df[i])
        return res


def estimate_means(mod: Model, method='ML', solver='SLSQP', 
                   pvals=False, ret_opt=False):
    """
    Estimate means for meanstructure-free model.

    Mean estimation is performed via fitting mean components to data given
    known covariance matrix Sigma. In case of ML/GLS, it's equivalent to
    minimizing Mahalanobis distance with respect to mean components. Usually,
    output variables estimates are really close to just data mean estimates.
    Parameters
    ----------
    mod : Model
        Model.
    method : str, optional
        Can 'ML'/'GLS', 'ULS'. The default is 'ML'.
    solver : str, optional
        Solver to use. The default is 'SLSQP'.
    pvals : bool, optional
        If True, then p-values are returned. They posses no information for
        most purposes here and usually are close to 1.0. The default is False.
    ret_opt : bool, optional
        If True, otpimizer info is also returned. The default is False.

    Returns
    -------
    pd.DataFrame
        Information on mean components.
    SolverResult, optional
        Information on optimization result.
    """
    me = MeanEstimator(mod)
    res = me.fit(method, solver)
    df = list()
    i = 0
    if pvals:
        se = calc_se(me)
        zv = calc_zvals(me, std_errors=se)
        pv = calc_pvals(me, z_scores=zv)
    else:
        se = np.zeros_like(res.x)
        zv = se
        pv = zv
    for v in mod.vars['inner']:
        if v in mod.vars['observed']:
            df.append((v, '~', '1', res.x[i], se[i], zv[i], pv[i]))
            i += 1
    for v in mod.vars['observed']:
        if v in mod.vars['_output']:
            df.append((v, '~', '1', res.x[i], se[i], zv[i], pv[i]))
            i += 1
    df = pd.DataFrame(df, columns=['lval', 'op', 'rval', 'Estimate',
                                   'Std. Err', 'z-value', 'p-value'])
    if not pvals:
        df = df.drop(['Std. Err', 'z-value', 'p-value'], axis=1)
    return df if not ret_opt else (df, res)



"""semopy: Structural Equation Modeling Optimization in Python"""

name = "semopy"
__version__ = "2.3.0"
__author__ = "Georgy Meshcheryakov"


from .model_generalized_effects import ModelGeneralizedEffects
from .model_effects import ModelEffects
from .model_means import ModelMeans
from .regularization import create_regularization
from .stats import calc_stats, gather_statistics
from .means import estimate_means
from .plot import semplot
from .model import Model
from .unbias import bias_correction
from .optimizer import Optimizer
from .report import report
from . import model_generation
from . import effects 
from . import multigroup
from . import examples
from . import utils
from . import efa
